"""Class API example cookbook with custom lock args."""
import argparse

from spicerack.cookbook import CookbookBase, CookbookRunnerBase, LockArgs


class ExampleCookbook(CookbookBase):
    """Cookbook that overrides the lock arguments."""

    def argument_parser(self):
        """As defined by the parent class."""
        return argparse.ArgumentParser("Argparse")

    def get_runner(self, args):
        """As required by the parent class."""
        return ExampleRunner()


class ExampleRunner(CookbookRunnerBase):
    """As required by spicerack._cookbook."""

    @property
    def lock_args(self):
        """Set the lock dynamically."""
        return LockArgs(suffix="test", concurrency=1, ttl=30)

    def run(self):
        """As required by the parent class."""



"""Netbox module."""
import logging
from ipaddress import IPv4Interface, IPv6Interface, ip_interface
from typing import Any, Optional, Union

import pynetbox
from requests.exceptions import RequestException
from wmflib.requests import http_session

from spicerack.decorators import retry
from spicerack.exceptions import SpicerackError

MANAGEMENT_IFACE_NAME: str = "mgmt"
"""The interface name used in Netbox for the OOB network."""
SERVER_ROLE_SLUG: str = "server"
"""Netbox role to identify servers."""
logger = logging.getLogger(__name__)


class NetboxError(SpicerackError):
    """General errors raised by this module."""


class NetboxAPIError(NetboxError):
    """Usually a wrapper for pynetbox.RequestError, errors that occur when accessing the API."""


class NetboxHostNotFoundError(NetboxError):
    """Raised when a host is not found for an operation."""


class NetboxScriptError(NetboxError):
    """Raised when a Netbox script doesn't run properly."""


class Netbox:
    """Class which wraps Netbox API operations."""

    def __init__(self, url: str, token: str, *, dry_run: bool = True):
        """Create Netbox instance.

        Arguments:
            url: The Netbox top level URL (with scheme and port if necessary).
            token: A Netbox API token.
            dry_run: set to False to cause writes to Netbox to occur.

        """
        self._api = pynetbox.api(url, token=token, threading=True)
        self._api.http_session = http_session(".".join((self.__module__, self.__class__.__name__)))
        self._dry_run = dry_run

    @property
    def api(self) -> pynetbox.api:
        """Get the pynetbox instance to interact directly with Netbox APIs.

        Caution:
            When feasible use higher level functionalities.

        """
        return self._api

    def _get_device(self, name: str) -> pynetbox.core.response.Record:
        """Get a device (dcim.devices) object.

        Arguments:
            name: the name of the device to get.

        Raises:
            spicerack.netbox.NetboxAPIError: on API error.
            spicerack.netbox.NetboxError: on parameter error.
            spicerack.netbox.NetboxHostNotFoundError: if the device is not found.

        """
        try:
            device = self._api.dcim.devices.get(name=name)
        except pynetbox.RequestError as ex:
            raise NetboxAPIError("Error retrieving Netbox device") from ex

        if device is None:
            raise NetboxHostNotFoundError(name)

        return device

    def _get_virtual_machine(self, hostname: str) -> pynetbox.core.response.Record:
        """Get a virtual machine (virtualization.virtual_machine) object.

        Arguments:
            hostname: the name of the host to get.

        Raises:
            spicerack.netbox.NetboxAPIError: on API error.
            spicerack.netbox.NetboxError: on parameter error.
            spicerack.netbox.NetboxHostNotFoundError: if the host is not found.

        """
        try:
            host = self._api.virtualization.virtual_machines.get(name=hostname)
        except pynetbox.RequestError as ex:
            raise NetboxAPIError("Error retrieving Netbox VM") from ex

        if host is None:
            raise NetboxHostNotFoundError

        return host

    def get_server(self, hostname: str) -> "NetboxServer":
        """Return a NetboxServer instance for the given hostname.

        Arguments:
            hostname: the device hostname.

        Raises:
            spicerack.netbox.NetboxHostNotFoundError: if the device can't be found among physical or virtual devices.
            spicerack.netbox.NetboxError: if the device is not a server.

        """
        try:
            server = self._get_device(hostname)
        except NetboxHostNotFoundError:
            server = self._get_virtual_machine(hostname)

        return NetboxServer(api=self._api, server=server, dry_run=self._dry_run)

    def run_script(self, name: str, *, commit: bool = False, params: dict[str, Any]) -> list:
        """Run a Netbox script and wait for its output.

        Arguments:
            name: Full name of the script to run (eg. import_server_facts.ImportPuppetDB).
            commit: save the script actions in the Netbox DB.
            params: script parameters (passed as POST data).

        Returns:
            The script execution logs in a list format.

        Raises:
            spicerack.netbox.NetboxScriptError: If the script coudn't be ran or its result fetched.

        """
        # Apparently pynetbox doesn't allow to execute a Netbox script
        url = self._api.extras.scripts.get(name).url
        headers = {"Authorization": f"Token {self._api.token}"}
        if self._dry_run and commit:
            logger.info("Forcing commit = False as running in DRY-RUN")
            commit = False
        data = {"data": params, "commit": int(commit)}

        script_http_session = http_session(".".join((self.__module__, self.__class__.__name__)), timeout=(5.0, 30.0))

        @retry(tries=30, backoff_mode="constant", exceptions=(ValueError, RequestException))
        def _poll_netbox_job(url: str) -> list:
            """Poll Netbox to get the result of the script run."""
            result = script_http_session.get(url, headers=headers)
            result.raise_for_status()
            data = result.json()["data"]
            if data is None:
                raise ValueError(f"No data from job result {url}")
            return data["log"]

        result = None
        try:
            result = script_http_session.post(url, headers=headers, json=data)
            result.raise_for_status()
            logger.debug("Started Netbox script %s, waiting for results.", name)
        except RequestException as e:
            raise NetboxScriptError(f"Failed to start Netbox script {name}") from e
        job_url = result.json()["result"]["url"]
        try:
            return _poll_netbox_job(job_url)
        except (ValueError, RequestException) as e:
            raise NetboxScriptError(f"Failed to get Netbox script results from {job_url}") from e


class NetboxServer:
    """Represent a Netbox device of role server or a virtual machine."""

    allowed_status_transitions: dict[str, tuple[str, ...]] = {
        "spare": ("planned", "failed", "decommissioned"),
        "planned": ("active", "failed", "decommissioned"),
        "failed": ("spare", "planned", "active", "decommissioned"),
        "active": ("failed", "decommissioned"),
        "decommissioned": ("planned", "spare"),
    }
    """Allowed transition between Netbox statuses.
    See https://wikitech.wikimedia.org/wiki/Server_Lifecycle#/media/File:Server_Lifecycle_Statuses.png"""

    def __init__(
        self,
        *,
        api: pynetbox.api,
        server: Union[pynetbox.models.dcim.Devices, pynetbox.models.virtualization.VirtualMachines],
        dry_run: bool = True,
    ):
        """Initialize the instance.

        Arguments:
            api: the API instance to connect to Netbox.
            server: the server object.

        Raises:
            spicerack.netbox.NetboxError: if the device is not of type server.

        """
        self._server = server
        self._api = api
        self._dry_run = dry_run
        self._cached_mgmt_fqdn = ""  # Cache the management interface as it would require an API call each time

        role = server.role.slug if self.virtual else server.device_role.slug
        if role != SERVER_ROLE_SLUG:
            raise NetboxError(f"Object of type {type(server)} has invalid role {role}, only server is allowed")

    @property
    def virtual(self) -> bool:
        """Getter to check if the server is physical or virtual.

        Returns:
            :py:data:`True` if the server is virtual, :py:data:`False` if physical.

        """
        return not hasattr(self._server, "rack")

    @property
    def name(self) -> str:
        """Get the server name. Set the server name, primary IP DNS name, management IP DNS name.

        Modifying its value can be done only on physical devices.

        Arguments:
            new: the new name for the host.

        Raises:
            spicerack.netbox.NetboxError: if trying to set it on a virtual device or there is an issue renaming.

        """
        return self._server.name

    @name.setter
    def name(self, new: str) -> None:
        """Get and set the server name. See the getter docstring for info.

        Arguments:
            new: the new name for the host.

        Raises:
            spicerack.netbox.NetboxError: if trying to set it on a virtual device or there is an issue renaming.

        """
        if self.virtual:
            raise NetboxError(
                f"Server {self._server.name} is a virtual machine, chaging the name is only for physical servers."
            )

        current = self._server.name
        if current == new:
            logger.debug("Current name is already %s", current)
            return
        if self._dry_run:
            logger.info("Skipping Netbox name change from %s to %s in DRY-RUN.", current, new)
            return

        self._server.name = new
        if self._server.save():
            logger.debug("Updated Netbox name from %s to %s", current, new)
        else:
            # See https://github.com/netbox-community/pynetbox/issues/586
            raise NetboxError(f"Name change for {current} didn't get applied by Netbox.")

        # Update the FQDNs
        self.fqdn = new + "." + self.fqdn.split(".", 1)[1]
        self.mgmt_fqdn = new + "." + self.mgmt_fqdn.split(".", 1)[1]

    @property
    def status(self) -> str:
        """Get and set the server status.

        Modifying its value can be done only on physical devices and only between allowed transitions.

        The allowed transitions are defined in :py:data:`spicerack.netbox.Netbox.allowed_status_transitions`.

        Arguments:
            value: the name of the status to be set. It will be lower cased automatically.

        Raises:
            spicerack.netbox.NetboxError: if trying to set it on a virtual device or the status transision is not
                allowed.

        """
        return self._server.status.value

    @status.setter
    def status(self, value: str) -> None:
        """Set the device status. See the getter docstring for info.

        Arguments:
            value: the name of the status to be set. It will be lower cased automatically.

        Raises:
            spicerack.netbox.NetboxError: if used on a virtual device or the status transision is not allowed.

        """
        if self.virtual:
            raise NetboxError(
                f"Server {self._server.name} is a virtual machine, its Netbox status is automatically synced from "
                f"Ganeti."
            )

        current = self._server.status.value
        new = value.lower()
        allowed_transitions = NetboxServer.allowed_status_transitions.get(current, ())
        if new not in allowed_transitions:
            raise NetboxError(
                f"Forbidden Netbox status transition between {current} and {new} for device {self._server.name}. "
                f"Possible values are: {allowed_transitions}"
            )

        if self._dry_run:
            logger.info(
                "Skipping Netbox status change from %s to %s for device %s in DRY-RUN.", current, new, self._server.name
            )
            return

        self._server.status = value
        self._server.save()
        logger.debug("Updated Netbox status from %s to %s for device %s", current, new, self._server.name)

    def _set_primary_ip(self, value: Union[str, IPv4Interface, IPv6Interface], version: int) -> None:
        """Abstraction function to set a v4 or v6 IP on a host.

        Arguments:
            value: an IPv4 or IPv6 IP in CIDR notation.

        Raises:
            spicerack.netbox.NetboxError: if the provided IP is not valid.

        """
        try:
            ip = ip_interface(value)
        except ValueError as exc:
            raise NetboxError(f"{value} is not a valid IP in the CIDR notation.") from exc
        if ip.version != version:
            raise NetboxError(f"{value} is not an IPv{version}")
        primary_ip = getattr(self._server, f"primary_ip{ip.version}")
        if not primary_ip:
            raise NetboxError(f"No existing primary IPv{version} for {self._server.name}.")
        current = primary_ip.address
        if self._api.ipam.ip_addresses.count(address=str(ip)):
            raise NetboxError(f"{ip} is already in use.")
        if self._dry_run:
            logger.info(
                "Skipping Netbox primary IPv%d change from %s to %s for device %s in DRY-RUN.",
                ip.version,
                current,
                value,
                self._server.name,
            )
            return

        primary_ip.address = str(ip)
        primary_ip.save()
        logger.debug(
            "Updated Netbox primary IPv%d from %s to %s for device %s",
            ip.version,
            current,
            value,
            self._server.name,
        )

    @property
    def primary_ip4_address(self) -> Optional[IPv4Interface]:
        """Get and set the server primary IPv4 address.

        And not the Netbox ipam.IpAddresses object.

        Arguments:
            value: the new IPv4 (CIDR) to be set.

        """
        if self._server.primary_ip4:
            return IPv4Interface(self._server.primary_ip4.address)
        return None

    @primary_ip4_address.setter
    def primary_ip4_address(self, value: Union[str, IPv4Interface]) -> None:
        """Set the server primary IPv4 address. See the getter docstring for info.

        And not the Netbox ipam.IpAddresses object.

        Arguments:
            value: the new IPv4 (CIDR) to be set.

        """
        self._set_primary_ip(value, 4)

    @property
    def primary_ip6_address(self) -> Optional[IPv6Interface]:
        """Get and set the server primary IPv6 address.

        And not the Netbox ipam.IpAddresses object.

        Arguments:
            value: the new IPv6 (CIDR) to be set.

        """
        if self._server.primary_ip6:
            return IPv6Interface(self._server.primary_ip6.address)
        return None

    @primary_ip6_address.setter
    def primary_ip6_address(self, value: Union[str, IPv6Interface]) -> None:
        """Set the server primary IPv6 address. See the getter docstring for info.

        And not the Netbox ipam.IpAddresses object.

        Arguments:
            value: the new IPv6 (CIDR) to be set.

        """
        self._set_primary_ip(value, 6)

    def _find_primary_switch_iface(self) -> pynetbox.core.response.Record:
        """Returns the switch side interface connected to the device's primary interface.

        Returns:
            A Netbox interface object of the switch side connected to the device's primary interface.

        Raises:
            spicerack.netbox.NetboxError: if used on a virtual device,
              the device primary interface is not connected, or the primary IP is not linked to an interface.

        """
        # TODO: in the future find another way than requiring a primary IP to find the primary interface
        if self.virtual:
            raise NetboxError("Server is a virtual machine, can't return a switch interface.")
        primary_ip = self._server.primary_ip
        if not primary_ip:
            raise NetboxError("No primary IP, needed to find the primary interface.")
        netbox_iface = primary_ip.assigned_object
        if not netbox_iface:
            raise NetboxError("Primary IP not assigned to an interface.")
        netbox_switch_iface = netbox_iface.connected_endpoint
        if not netbox_switch_iface:
            raise NetboxError("Primary interface not connected.")
        return netbox_switch_iface

    @property
    def access_vlan(self) -> str:
        """Get and set the server access vlan.

        Can be done only on physical devices, requires the device to have a primary IP.

        Arguments:
            value: the name of the vlan to be set.

        """
        netbox_switch_iface = self._find_primary_switch_iface()
        if netbox_switch_iface.untagged_vlan:
            return netbox_switch_iface.untagged_vlan.name
        return ""

    @access_vlan.setter
    def access_vlan(self, value: str) -> None:
        """Set the device access vlan. See the getter docstring for info.

        Arguments:
            value: the name of the access vlan to be set.

        Raises:
            spicerack.netbox.NetboxError: if used on a virtual device,
              the device primary interface is not connected or the vlan doesn't exist.

        """
        netbox_switch_iface = self._find_primary_switch_iface()
        new_vlan = self._api.ipam.vlans.get(name=value, status="active")
        if not new_vlan:
            raise NetboxError(f"Failed to find an active VLAN with name {value}")

        current = netbox_switch_iface.untagged_vlan.name if netbox_switch_iface.untagged_vlan else "NOT SET"
        if self._dry_run:
            logger.info(
                "Skipping Netbox update of switchport access vlan from %s to %s for device %s in DRY-RUN.",
                current,
                value,
                self._server.name,
            )
            return
        netbox_switch_iface.untagged_vlan = new_vlan
        netbox_switch_iface.save()
        logger.debug(
            "Updated Netbox switchport access vlan from %s to %s for device %s", current, value, self._server.name
        )

    @property
    def fqdn(self) -> str:
        """Get and set the device primary IPs FQDN if one is already set.

        Notes:
            If the FQDN for any of the primary IPs is not set it will not be updated.
            This is to prevent setting a IPv6 AAAA record by accident.

        Arguments:
            value: the new FQDN for the host.

        Raises:
            spicerack.netbox.NetboxError: if the server has no FQDN defined in Netbox.

        """
        # Until https://phabricator.wikimedia.org/T253173 is fixed we can't use the primary_ip attribute
        for attr_name in ("primary_ip4", "primary_ip6"):
            address = getattr(self._server, attr_name)
            if address is not None and address.dns_name:
                return address.dns_name

        raise NetboxError(f"Server {self._server.name} does not have any primary IP with a DNS name set.")

    @fqdn.setter
    def fqdn(self, value: str) -> None:
        """Get and set the device primary IPs FQDN if one is already set.

        Notes:
            If the FQDN for any of the primary IPs is not set it will not be updated.
            This is to prevent setting a IPv6 AAAA record by accident.

        Arguments:
            value: the new FQDN for the host.

        Raises:
            spicerack.netbox.NetboxError: if trying to set it on a virtual device
                                          or if the server has no FQDN defined in Netbox.

        """
        if self.virtual:
            raise NetboxError(
                f"Server {self._server.name} is a virtual machine, changing the FQDN is only for physical servers."
            )
        for attr_name in ("primary_ip4", "primary_ip6"):
            address = getattr(self._server, attr_name)
            if address is not None and address.dns_name:
                if address.dns_name == value:
                    logger.debug("Current dns_name is already %s", value)
                    continue
                address.dns_name = value
                if not address.save():
                    raise NetboxError(f"Spicerack was not able to update the {attr_name} FQDN for {self._server.name}.")
                logger.debug("Updated %s dns_name to %s", attr_name, value)

    @property
    def mgmt_fqdn(self) -> str:
        """Get and set the management FQDN of the device.

        Arguments:
            value: the new FQDN for the host.

        Raises:
            spicerack.netbox.NetboxError: for virtual servers or the server has no management FQDN defined in Netbox.

        """
        if self.virtual:
            raise NetboxError(f"Server {self._server.name} is a virtual machine, does not have a management address.")

        if self._cached_mgmt_fqdn:
            return self._cached_mgmt_fqdn

        address = self._api.ipam.ip_addresses.get(device=self._server.name, interface=MANAGEMENT_IFACE_NAME)
        # TODO: check also that address.assigned_object.mgmt_only is True if it will not generate anymore an additional
        #       API call to Netbox or the Netbox API become more efficient.
        if address is not None and address.dns_name:
            self._cached_mgmt_fqdn = address.dns_name
            return self._cached_mgmt_fqdn

        raise NetboxError(f"Server {self._server.name} has no management interface with a DNS name set.")

    @mgmt_fqdn.setter
    def mgmt_fqdn(self, value: str) -> None:
        """Get and set the management FQDN of the device.

        Arguments:
            value: the new FQDN for the host.

        Raises:
            spicerack.netbox.NetboxError: if trying to set it on a virtual device or can't find the management IP.

        """
        if self.virtual:
            raise NetboxError(
                f"Server {self._server.name} is a virtual machine, "
                "changing the mgmt FQDN is only for physical servers."
            )
        address = self._api.ipam.ip_addresses.get(device=self._server.name, interface=MANAGEMENT_IFACE_NAME)
        # TODO: see the getter TODO
        if address is not None:
            if address.dns_name == value:
                self._cached_mgmt_fqdn = value
                logger.debug("Current dns_name is already %s", value)
                return
            address.dns_name = value
            if not address.save():
                raise NetboxError(f"Spicerack was not able to update the mgmt_fqdn for {self._server.name}.")
            self._cached_mgmt_fqdn = value
            logger.debug("Updated mgmt FQDN to %s", value)

    @property
    def asset_tag_fqdn(self) -> str:
        """Return the management FQDN for the asset tag of the device.

        Raises:
            spicerack.netbox.NetboxError: for virtual servers or the server has no management FQDN defined in Netbox.

        """
        parts = self.mgmt_fqdn.split(".")
        parts[0] = self._server.asset_tag.lower()
        return ".".join(parts)

    def as_dict(self) -> dict:
        """Return a dict containing details about the server."""
        ret = dict(self._server)
        ret["is_virtual"] = self.virtual

        return ret



"""locking module tests."""
import json
import logging
import re
import uuid
from datetime import datetime, timedelta
from unittest import mock

import pytest

from spicerack import locking
from spicerack.tests import get_fixture_path

CREATED_DATETIME = datetime(2023, 1, 1, 12, 34, 56, 123456)
CONCURRENT_LOCK_ARGS = {"concurrency": 2, "owner": "user@host [123]", "ttl": 60}
SAMPLE_UUID = "4734a179-0062-436d-a9d5-4cf0c24a12d7"
KEY_LOCKS_JSON = f"""
{{
    "{SAMPLE_UUID}": {{
        "concurrency": 2, "created": "2023-01-01 12:34:56.123456", "owner": "user@host [123]", "ttl": 60
        }}
}}
"""


@pytest.mark.parametrize("stem", ("config", "non-existent"))
def test_get_lock_instance(stem, monkeypatch):
    """It should return a Lock instance."""
    monkeypatch.setenv("USER", "")
    config_file = get_fixture_path("locking", "config.yaml")
    with mock.patch("spicerack.locking.Path") as mocked_path:
        mocked_path.return_value = config_file.with_stem(stem)
        lock = locking.get_lock_instance(
            config_file=config_file,
            prefix="cookbooks",
            owner="user@host [123]",
            dry_run=False,
        )
        mocked_path.assert_called_once_with("~/.etcdrc")

    assert isinstance(lock, locking.Lock)


def test_get_lock_instance_dry_run():
    """It should return a NoLock instance."""
    lock = locking.get_lock_instance(config_file=None, prefix="modules", owner="user@host [123]", dry_run=True)
    assert isinstance(lock, locking.NoLock)


def test_get_lock_instance_invalid_prefix():
    """It should raise a LockError exception if the prefix is invalid."""
    with pytest.raises(locking.LockError, match="Invalid prefix invalid, must be one of:"):
        locking.get_lock_instance(config_file=None, prefix="invalid", owner="user@host [123]", dry_run=True)


class TestLock:
    """Test the Lock class."""

    # Adding an additional mock here for Lock or trying to mock the whole etcd module doens't work as expected
    @mock.patch("spicerack.locking.etcd.Client", autospec=True)
    def setup_method(self, _, mocked_client):
        """Initialize the test environment."""
        # pylint: disable=attribute-defined-outside-init
        self.mocked_client = mocked_client.return_value
        self.mocked_client.read_timeout = 10
        self.full_key = "/spicerack/locks/cookbooks/key"
        self.lock = locking.Lock(config={}, prefix="cookbooks", owner="user@host [123]", dry_run=False)
        self.lock_dry_run = locking.Lock(config={}, prefix="modules", owner="user@host [123]", dry_run=True)
        self.mocked_client.reset_mock()

    def test_init_ok(self):
        """It should create an instance of the Lock class."""
        assert isinstance(self.lock, locking.Lock)

    def test_init_fail(self):
        """It should raise a LockError exception if the prefix is not valid."""
        with pytest.raises(locking.LockError, match="Invalid prefix invalid, must be one of"):
            locking.Lock(config={}, prefix="invalid", owner="user@host [123]", dry_run=True)

    def test_get_ok(self):
        """It should return the existing key if present."""
        self.mocked_client.read.return_value.value = KEY_LOCKS_JSON
        obj = self.lock.get("key")
        assert isinstance(obj, locking.KeyLocks)
        assert obj.key == self.full_key
        assert obj.locks[SAMPLE_UUID].concurrency == 2

    def test_get_missing(self):
        """It should return an empty KeyLocks instance if the key is missing."""
        self.mocked_client.read.side_effect = KeyError
        obj = self.lock.get("key")
        assert isinstance(obj, locking.KeyLocks)
        assert obj.key == self.full_key
        assert not obj.locks

    @pytest.mark.parametrize(
        "name, error",
        (
            ("", "The lock name cannot be empty"),
            ("invalid/name", "The lock name cannot contain '/', got: invalid/name"),
            ("key", "Locks with prefix 'modules' must have names starting with 'spicerack.', got: key"),
        ),
    )
    def test_get_invalid_name(self, name, error):
        """It should raise an InvalidLockError exception if the name contains a slash."""
        self.mocked_client.read.return_value.value = KEY_LOCKS_JSON
        with pytest.raises(locking.InvalidLockError, match=re.escape(error)):
            self.lock_dry_run.get(name)

    def test_get_fail(self):
        """It should raise a LockError exception if there is a failure."""
        self.mocked_client.read.side_effect = locking.etcd.EtcdException
        with pytest.raises(locking.LockError, match=f"Failed to get key {self.full_key}"):
            self.lock.get("key")

    @mock.patch("spicerack.locking.etcd.Lock", autospec=True)
    @mock.patch("uuid.uuid4", return_value="b8dfe73c-0d2c-4406-b783-f7118911014f")
    def test_acquired_ok(self, mocked_uuid, mocked_lock, caplog):
        """It should acquire the lock on the backend, give back control and then release it."""
        mocked_lock.return_value.is_acquired = True
        lock_payload = KEY_LOCKS_JSON.replace(SAMPLE_UUID, mocked_uuid())
        type(self.mocked_client.read.return_value).value = mock.PropertyMock(side_effect=[KEY_LOCKS_JSON, lock_payload])
        executed = False
        mocked_uuid.reset_mock()

        with caplog.at_level(logging.INFO):
            with self.lock.acquired("key", concurrency=2, ttl=60):
                executed = True

        mocked_uuid.assert_called_once()
        self.mocked_client.set.assert_called_once()
        self.mocked_client.delete.assert_called_once_with(self.full_key)
        assert self.mocked_client.set.call_args.args[0] == self.full_key
        saved_payload = json.loads(self.mocked_client.set.call_args.args[1])
        assert len(saved_payload) == 1
        assert saved_payload[mocked_uuid()]["concurrency"] == 2
        assert saved_payload[mocked_uuid()]["owner"] == "user@host [123]"
        assert saved_payload[mocked_uuid()]["ttl"] == 60
        assert executed
        assert f"Acquired lock for key {self.full_key}:" in caplog.text
        assert f"Released lock for key {self.full_key}:" in caplog.text

    def test_acquire_dry_run_ok(self, caplog):
        """It should mimic acquiring the lock on the backend without actually writing to it."""
        self.mocked_client.read.return_value.value = KEY_LOCKS_JSON
        with caplog.at_level(logging.INFO):
            self.lock_dry_run.acquire("spicerack.module.name", concurrency=2, ttl=60)

        self.mocked_client.set.assert_not_called()
        assert "Skipping lock acquire/release in DRY-RUN mode" in caplog.text
        assert "/spicerack/locks/modules/spicerack.module.name" in caplog.text

    @mock.patch("spicerack.locking.etcd.Lock", autospec=True)
    def test_acquire_ok(self, mocked_lock, caplog):
        """It should acquire the lock on the backend."""
        mocked_lock.return_value.is_acquired = True
        self.mocked_client.read.return_value.value = KEY_LOCKS_JSON
        with caplog.at_level(logging.INFO):
            lock_id = self.lock.acquire("key", concurrency=2, ttl=60)

        uuid.UUID(lock_id)  # Raises if it's not a valid UUID
        self.mocked_client.set.assert_called_once()
        assert f"Acquired lock for key {self.full_key}:" in caplog.text

    @mock.patch("spicerack.locking.etcd.Lock", autospec=True)
    @mock.patch("time.sleep", return_value=None)
    def test_acquire_fail_release_ok(self, mocked_sleep, mocked_lock, caplog):
        """It should raise a EtcdLockUnavailableError exception if unable to acquire the lock."""
        mocked_lock.return_value.acquire.side_effect = locking.etcd.EtcdException("##FAILED##")
        with caplog.at_level(logging.INFO):
            with pytest.raises(locking.EtcdLockUnavailableError, match="##FAILED##"):
                self.lock.acquire("key", concurrency=2, ttl=60)

        assert mocked_sleep.call_count == 14
        self.mocked_client.set.assert_not_called()
        assert "Acquired lock" not in caplog.text

    @mock.patch("spicerack.locking.etcd.Lock", autospec=True)
    @mock.patch("time.sleep", return_value=None)
    def test_acquire_fail_release_fail(self, mocked_sleep, mocked_lock, caplog):
        """It should raise a EtcdLockUnavailableError exception if unable to acquire the lock and log a warning."""
        mocked_lock.return_value.acquire.side_effect = locking.etcd.EtcdException("##FAILED##")
        mocked_lock.return_value.release.side_effect = locking.etcd.EtcdException("##FAILED_RELEASE##")
        mocked_lock.return_value.path = "##LOCK_PATH##"
        with caplog.at_level(logging.INFO):
            with pytest.raises(locking.EtcdLockUnavailableError, match="##FAILED##"):
                self.lock.acquire("key", concurrency=2, ttl=60)

        assert mocked_sleep.call_count == 14
        self.mocked_client.set.assert_not_called()
        assert "Acquired lock" not in caplog.text
        assert "Failed to release etcd attempted lock queued" in caplog.text
        assert "##FAILED_RELEASE##" in caplog.text

    @mock.patch("spicerack.locking.etcd.Lock", autospec=True)
    @mock.patch("time.sleep", return_value=None)
    def test_acquire_not_acquired_release_ok(self, mocked_sleep, mocked_lock, caplog):
        """It should raise a EtcdLockUnavailableError exception if the lock is already taken."""
        mocked_lock.return_value.is_acquired = False
        with caplog.at_level(logging.INFO):
            with pytest.raises(locking.EtcdLockUnavailableError, match="Lock already taken"):
                self.lock.acquire("key", concurrency=2, ttl=60)

        assert mocked_sleep.call_count == 14
        self.mocked_client.set.assert_not_called()
        assert "Acquired lock" not in caplog.text

    @mock.patch("spicerack.locking.etcd.Lock", autospec=True)
    @mock.patch("time.sleep", return_value=None)
    def test_acquire_not_acquired_release_fail(self, mocked_sleep, mocked_lock, caplog):
        """It should raise a EtcdLockUnavailableError exception if the lock is already taken."""
        mocked_lock.return_value.is_acquired = False
        mocked_lock.return_value.release.side_effect = locking.etcd.EtcdException("##FAILED_RELEASE##")
        mocked_lock.return_value.path = "##LOCK_PATH##"
        with caplog.at_level(logging.INFO):
            with pytest.raises(locking.EtcdLockUnavailableError, match="Lock already taken"):
                self.lock.acquire("key", concurrency=2, ttl=60)

        assert mocked_sleep.call_count == 14
        self.mocked_client.set.assert_not_called()
        assert "Acquired lock" not in caplog.text
        assert "Failed to release etcd lock queued" in caplog.text
        assert "##FAILED_RELEASE##" in caplog.text

    @mock.patch("spicerack.locking.etcd.Lock", autospec=True)
    def test_release_ok(self, mocked_lock, caplog):
        """It should acquire and then delete the lock on the backend."""
        mocked_lock.return_value.is_acquired = True
        self.mocked_client.read.return_value.value = KEY_LOCKS_JSON
        lock_id = self.lock.acquire("key", concurrency=2, ttl=60)
        self.mocked_client.reset_mock()
        self.mocked_client.read.return_value.value = KEY_LOCKS_JSON.replace(SAMPLE_UUID, lock_id)

        with caplog.at_level(logging.INFO):
            self.lock.release("key", lock_id)

        self.mocked_client.delete.assert_called_once_with(self.full_key)
        assert f"Released lock for key {self.full_key}:" in caplog.text

    @mock.patch("spicerack.locking.etcd.Lock", autospec=True)
    def test_release_remaining_ok(self, mocked_lock, caplog):
        """It should acquire and then release the lock on the backend keeping the existing one."""
        locks = json.loads(KEY_LOCKS_JSON)
        locks[SAMPLE_UUID]["created"] = str(datetime.utcnow())
        mocked_lock.return_value.is_acquired = True
        self.mocked_client.read.return_value.value = json.dumps(locks)
        lock_id = self.lock.acquire("key", concurrency=2, ttl=60)
        self.mocked_client.reset_mock()
        locks[lock_id] = locks[SAMPLE_UUID].copy()
        locks[lock_id]["created"] = str(datetime.utcnow())
        self.mocked_client.read.return_value.value = json.dumps(locks)

        with caplog.at_level(logging.INFO):
            self.lock.release("key", lock_id)

        del locks[lock_id]
        self.mocked_client.set.assert_called_once_with(self.full_key, json.dumps(locks))
        assert f"Released lock for key {self.full_key}:" in caplog.text

    @mock.patch("spicerack.locking.etcd.Lock", autospec=True)
    @mock.patch("time.sleep", return_value=None)
    def test_release_fail(self, mocked_sleep, mocked_lock, caplog):
        """It should not raise an exception if the lock release fails."""
        mocked_lock.return_value.is_acquired = False
        with caplog.at_level(logging.ERROR):
            self.lock.release("key", SAMPLE_UUID)

        assert mocked_sleep.call_count == 14
        self.mocked_client.set.assert_not_called()
        assert f"Failed to release lock for key {self.full_key} and ID {SAMPLE_UUID}: Lock already taken" in caplog.text

    @mock.patch("spicerack.locking.etcd.Lock", autospec=True)
    def test_release_missing(self, mocked_lock, caplog):
        """It should not write to the backend if the lock has been already removed."""
        mocked_lock.return_value.is_acquired = True
        self.mocked_client.read.return_value.value = "{}"
        with caplog.at_level(logging.WARNING):
            self.lock.release("key", "missing")

        self.mocked_client.set.assert_not_called()
        assert "Released lock for key" not in caplog.text
        assert (
            f"Lock for key {self.full_key} and ID missing not found. Unable to release it. Was expired?" in caplog.text
        )

    def test_release_dry_run_ok(self, caplog):
        """It should mimic releasing the lock on the backend without actually writing to it."""
        self.mocked_client.read.return_value.value = KEY_LOCKS_JSON
        with caplog.at_level(logging.INFO):
            self.lock_dry_run.release("spicerack.module.name", SAMPLE_UUID)

        self.mocked_client.set.assert_not_called()
        assert "Skipping lock acquire/release in DRY-RUN mode" in caplog.text


class TestNoLock:
    """Test the NoLock class."""

    def setup_method(self):
        """Initialize the test environment."""
        # pylint: disable=attribute-defined-outside-init
        self.lock = locking.NoLock(config={}, prefix="prefix", owner="user@host [123]", dry_run=False)

    def test_get(self):
        """It should return an empty KeyLocks instance."""
        key_locks = self.lock.get("name")
        assert key_locks.key == "name"
        assert not key_locks.locks

    def test_acquired(self):
        """It should just give back control to the user."""
        with self.lock.acquired():
            return True

    def test_acquire(self):
        """It should just do nothing and return an empty lock id."""
        lock_id = self.lock.acquire()
        assert lock_id == ""

    def test_release(self):
        """It should just do nothing."""
        self.lock.release("")


class TestKeyLocks:
    """Test the KeyLocks class."""

    def setup_method(self):
        """Initialize the test environment."""
        # pylint: disable=attribute-defined-outside-init
        self.key = "test.key"
        self.lock = locking.KeyLocks(key=self.key)
        self.concurrent_lock = locking.ConcurrentLock(
            uuid=SAMPLE_UUID, created=CREATED_DATETIME, **CONCURRENT_LOCK_ARGS
        )

    def test_from_json_ok(self):
        """It should create a KeyLocks instance from a JSON payload."""
        lock = locking.KeyLocks.from_json(self.key, KEY_LOCKS_JSON)
        assert isinstance(lock, locking.KeyLocks)
        assert lock.key == self.key
        assert lock.locks[SAMPLE_UUID].created == CREATED_DATETIME

    def test_from_json_fail(self):
        """It should raise a LockUnreadableError exception if unable to parse the JSON lock."""
        with pytest.raises(
            locking.LockUnreadableError, match=re.escape(f"Unable to JSON-deserialize the lock for key {self.key}")
        ):
            locking.KeyLocks.from_json(self.key, "invalid")

    def test_to_json_ok(self):
        """It should return the JSON-serialized object as a string."""
        # Prevent the date to get updated
        self.concurrent_lock.update_created = mock.Mock()
        self.lock.add(self.concurrent_lock)
        assert self.lock.to_json() == json.dumps(json.loads(KEY_LOCKS_JSON))  # Removes the pretty printing

    def test_to_json_fail(self):
        """It should raise a LockUnwritableError exception if unable to serialize the object."""
        self.concurrent_lock.uuid = uuid.uuid4()  # Replace it with an instance, not a string
        self.lock.add(self.concurrent_lock)
        with pytest.raises(locking.LockUnwritableError, match="Unable to JSON-serialize the current instance"):
            self.lock.to_json()

    def test_add_ok(self):
        """It should add the concurrent lock to the instance and update its created time."""
        self.lock.add(self.concurrent_lock)
        assert len(self.lock.locks) == 1
        lock = self.lock.locks[SAMPLE_UUID]
        assert lock.created > datetime.utcnow() - timedelta(seconds=10)
        assert lock == self.concurrent_lock

    def test_add_concurrency_already_reached(self):
        """It should raise a LockUnavailableError exception if there are already lock with reached concurrency."""
        lock1 = locking.ConcurrentLock(owner="##1##", concurrency=2, ttl=10)
        lock2 = locking.ConcurrentLock(owner="##2##", concurrency=2, ttl=10)
        self.lock.add(lock1)
        self.lock.add(lock2)
        expected = (
            f"for key {self.key}.\n"
            "There are already 2 concurrent locks and the concurrency allowed is 2:\n"
            f"    {lock1.created}: ##1##\n"
            f"    {lock2.created}: ##2##"
        )
        with pytest.raises(locking.LockUnavailableError, match=re.escape(expected)):
            self.lock.add(locking.ConcurrentLock(owner="##3##", concurrency=5, ttl=10))

    def test_add_removes_expired(self, caplog):
        """It should remove the expired locks if there is any."""
        kwargs = CONCURRENT_LOCK_ARGS.copy()
        kwargs["concurrency"] = 5
        for i in range(4):
            kwargs["owner"] = f"##{i}##"
            lock = locking.ConcurrentLock(**kwargs)
            self.lock.add(lock)

        for lock in self.lock.locks.values():
            lock.created = CREATED_DATETIME

        with caplog.at_level(logging.INFO):
            self.lock.add(locking.ConcurrentLock(**kwargs))

        assert caplog.text.count(f"Releasing expired lock for key {self.key}:") == 4
        for i in range(4):
            assert f"##{i}##" in caplog.text

    def test_add_duplicate_id(self):
        """It should raise a LockExistingError exception if there is already a lock with the same ID."""
        self.lock.add(self.concurrent_lock)
        with pytest.raises(
            locking.LockExistingError,
            match=re.escape(f"for key {self.key}. A lock with the same ID {SAMPLE_UUID} already exists:"),
        ):
            self.lock.add(self.concurrent_lock)

    def test_add_concurrency_reached(self):
        """It should raise a LockUnavailableError exception if there are already concurrency locks."""
        self.lock.add(self.concurrent_lock)
        kwargs = CONCURRENT_LOCK_ARGS.copy()
        kwargs["concurrency"] = 1
        with pytest.raises(
            locking.LockUnavailableError,
            match=re.escape(
                f"for key {self.key}. There are already 1 concurrent locks and a concurrency of 1 was requested"
            ),
        ):
            self.lock.add(locking.ConcurrentLock(**kwargs))

    def test_remove_ok(self):
        """It should remove the concurent lock from the current instance and return it."""
        self.lock.add(self.concurrent_lock)
        lock = self.lock.remove(self.concurrent_lock.uuid)
        assert lock == self.concurrent_lock
        assert not self.lock.locks

    def test_remove_missing(self, caplog):
        """It should return None and log a warning message."""
        self.lock.add(self.concurrent_lock)
        prev_len = len(self.lock.locks)
        with caplog.at_level(logging.WARNING):
            lock = self.lock.remove("non-existent")

        assert prev_len == len(self.lock.locks)
        assert lock is None
        assert f"Lock for key {self.key} and ID non-existent not found. Unable to release it." in caplog.text


class TestConcurrentLock:
    """Test the ConcurrentLock class."""

    def setup_method(self):
        """Initialize the test environment."""
        # pylint: disable=attribute-defined-outside-init
        self.uuid = str(uuid.uuid4())
        self.lock = locking.ConcurrentLock(created=CREATED_DATETIME, **CONCURRENT_LOCK_ARGS)

    @pytest.mark.parametrize(
        "kwargs, message",
        (
            (
                {"concurrency": -1, "owner": "user@host [123]", "ttl": 60},
                "argument 'concurrency' must be a non-negative integer, got -1",
            ),
            (
                {"concurrency": 1, "owner": "user@host [123]", "ttl": 0},
                "argument 'ttl' must be a positive integer, got 0",
            ),
            (
                {
                    "concurrency": 1,
                    "owner": "user@host [123]",
                    "ttl": 60,
                    "created": datetime.utcnow() + timedelta(seconds=60),
                },
                "argument 'created' cannot be in the future",
            ),
        ),
    )
    def test_init_fail(self, kwargs, message):
        """Instantiating a new ConcurrentLock object shuould fail if the parameters are not correct."""
        with pytest.raises(locking.InvalidLockError, match=re.escape(message)):
            locking.ConcurrentLock(**kwargs)

    def test_str(self):
        """Test the string representation of the object."""
        assert (
            str(self.lock)
            == "{'concurrency': 2, 'created': '2023-01-01 12:34:56.123456', 'owner': 'user@host [123]', 'ttl': 60}"
        )

    def test_update_created(self):
        """It should update the created property to now."""
        self.lock.update_created()
        assert self.lock.created >= datetime.utcnow() - timedelta(seconds=10)

    def test_to_dict(self):
        """If should return the dictionary custom representation of the object suitable for the locking backend."""
        assert self.lock.to_dict() == {"created": str(CREATED_DATETIME), **CONCURRENT_LOCK_ARGS}

    def test_from_dict_ok(self):
        """It should create an instance of ConcurrentLock from a dictionary."""
        lock = locking.ConcurrentLock.from_dict(self.uuid, {"created": str(CREATED_DATETIME), **CONCURRENT_LOCK_ARGS})
        assert isinstance(lock, locking.ConcurrentLock)
        assert lock.uuid == self.uuid
        assert lock.created == CREATED_DATETIME

    def test_from_dict_fail(self):
        """It should fail to create an instance of ConcurrentLock from a dictionary if the date cannot be parsed."""
        with pytest.raises(locking.InvalidLockError, match="key 'created' cannot be parsed as a datetime object"):
            locking.ConcurrentLock.from_dict(self.uuid, {"created": "invalid", **CONCURRENT_LOCK_ARGS})



"""Custom Spicerack type hints."""
from collections.abc import Sequence
from typing import TypeVar

from cumin import NodeSet

TypeHosts = TypeVar("TypeHosts", Sequence[str], NodeSet)
"""Custom type for hosts, can be a :py:class:`ClusterShell.NodeSet.NodeSet` or any sequence of :py:class:`str`."""



"""Vulture whitelist to avoid false positives."""


class Whitelist:
    """Helper class that allows mocking Python objects."""

    def __getattr__(self, _):
        """Mocking magic method __getattr__."""
        pass


whitelist_logging = Whitelist()
whitelist_logging.raiseExceptions

whitelist_ganeti = Whitelist()
whitelist_ganeti.Ganeti._http_session.auth

whitelist__menu = Whitelist()
whitelist__menu.CookbookItem._parse_args.parser.prog

# Needed because of https://github.com/jendrikseipp/vulture/issues/264
whitelist_dhcp = Whitelist()
whitelist_dhcp.DHCPConfOpt82.ipv4
whitelist_dhcp.DHCPConfOpt82.switch_hostname
whitelist_dhcp.DHCPConfOpt82.switch_iface
whitelist_dhcp.DHCPConfOpt82.vlan
whitelist_dhcp.DHCPConfOpt82.distro
whitelist_dhcp.DHCPConfOpt82.media_type
whitelist_dhcp.DHCPConfMac.ipv4
whitelist_dhcp.DHCPConfMac.distro
whitelist_dhcp.DHCPConfMac.media_type
whitelist_dhcp.DHCPConfMgmt.ipv4

whitelist_dnsdisc = Whitelist()
whitelist_dnsdisc.Discovery._resolvers.nameservers

whitelist_icinga = Whitelist()
whitelist_icinga.CommandFile.__new__
whitelist_icinga.IcingaStatus.CRITICAL
whitelist_icinga.IcingaStatus.UNKNOWN

whitelist_mysql = Whitelist()
whitelist_mysql.set_core_masters_readonly
whitelist_mysql.set_core_masters_readwrite

whitelist_redfish = Whitelist()
whitelist_redfish.ChassisResetPolicy.FORCE_RESTART
whitelist_redfish.ChassisResetPolicy.GRACEFUL_RESTART
whitelist_redfish.ChassisResetPolicy.GRACEFUL_SHUTDOWN
whitelist_redfish.DellSCPRebootPolicy.FORCED
whitelist_redfish.DellSCPRebootPolicy.GRACEFUL
whitelist_redfish.DellSCPPowerStatePolicy.OFF
whitelist_redfish.DellSCPTargetPolicy.BIOS
whitelist_redfish.DellSCPTargetPolicy.IDRAC
whitelist_redfish.DellSCPTargetPolicy.NIC
whitelist_redfish.DellSCPTargetPolicy.RAID
whitelist_redfish.DellSCP.comments

whitelist_remote = Whitelist()
whitelist_remote.execute.worker.progress_bars
whitelist_remote.execute.worker.reporter

# Needed because of https://github.com/jendrikseipp/vulture/issues/264
whitelist_service = Whitelist()
whitelist_service.Service.aliases
whitelist_service.Service.bgp
whitelist_service.Service.depool_threshold
whitelist_service.Service.encryption
whitelist_service.Service.httpbb_dir
whitelist_service.Service.lvs
whitelist_service.Service.lvs_class
whitelist_service.Service.monitors
whitelist_service.Service.page
whitelist_service.Service.probes
whitelist_service.Service.protocol
whitelist_service.Service.public_aliases
whitelist_service.Service.public_endpoint
whitelist_service.Service.scheduler
whitelist_service.Service.ipip_encapsulation
whitelist_service.ServiceMonitoring.check_command
whitelist_service.ServiceMonitoring.contact_group
whitelist_service.ServiceMonitoring.notes_url

whitelist_tests = Whitelist()
whitelist_tests.unit.test_confctl.TestConfctl.setup_method.backend
whitelist_tests.unit.test_dnsdisc.MockDnsResult.canonical_name
whitelist_tests.unit.test_dnsdisc.MockDnsResult.minimum_ttl
whitelist_tests.unit.test_netbox._netbox_host
whitelist_tests.unit.test_netbox._netbox_virtual_machine



"""Initialization tests."""

import logging
import sys
from collections import namedtuple
from importlib import import_module
from socket import gethostname
from unittest import mock

import pytest
from git import Repo
from requests import Session
from requests.auth import HTTPBasicAuth
from wmflib.actions import ActionsDict
from wmflib.config import load_yaml_config
from wmflib.dns import Dns
from wmflib.phabricator import Phabricator
from wmflib.prometheus import Prometheus, Thanos

from spicerack import Spicerack
from spicerack.administrative import Reason
from spicerack.alerting import AlertingHosts
from spicerack.alertmanager import Alertmanager, AlertmanagerHosts
from spicerack.apt import AptGetHosts
from spicerack.confctl import ConftoolEntity
from spicerack.debmonitor import Debmonitor
from spicerack.dhcp import DHCP
from spicerack.dnsdisc import Discovery
from spicerack.elasticsearch_cluster import ElasticsearchClusters
from spicerack.exceptions import SpicerackError
from spicerack.ganeti import Ganeti
from spicerack.icinga import IcingaHosts
from spicerack.ipmi import Ipmi
from spicerack.k8s import Kubernetes
from spicerack.kafka import Kafka
from spicerack.locking import Lock, NoLock
from spicerack.mediawiki import MediaWiki
from spicerack.mysql import Mysql
from spicerack.mysql_legacy import MysqlLegacy
from spicerack.netbox import Netbox, NetboxServer
from spicerack.peeringdb import PeeringDB
from spicerack.puppet import PuppetHosts, PuppetMaster, PuppetServer
from spicerack.redfish import RedfishDell, RedfishSupermicro
from spicerack.redis_cluster import RedisCluster
from spicerack.remote import Remote, RemoteError, RemoteHosts
from spicerack.reposync import RepoSync
from spicerack.service import Catalog
from spicerack.tests import SPICERACK_TEST_PARAMS, get_fixture_path
from spicerack.toolforge.etcdctl import EtcdctlController

MockedDnsSrv = namedtuple("MockedDnsSrv", ["target"])
MockedDnsAnswer = namedtuple("MockedDnsAnswer", ["ttl", "rrset"])


@mock.patch("wmflib.dns.resolver", autospec=True)
def test_spicerack(mocked_dns_resolver, monkeypatch):
    """An instance of Spicerack should allow to access all the library features."""
    monkeypatch.setenv("SUDO_USER", "user1")
    verbose = True
    dry_run = False
    proxy = "http://proxy.example.com:8080"
    spicerack = Spicerack(verbose=verbose, dry_run=dry_run, http_proxy=proxy, **SPICERACK_TEST_PARAMS)

    assert spicerack.verbose is verbose
    assert spicerack.dry_run is dry_run
    assert spicerack.username == "user1"
    assert spicerack.config_dir == get_fixture_path()
    assert spicerack.http_proxy == proxy
    assert spicerack.current_hostname == gethostname()
    assert spicerack.requests_proxies == {"http": proxy, "https": proxy}
    assert spicerack.authdns_servers == {"authdns1001.example.org": "10.0.0.1", "authdns2001.example.org": "10.0.0.2"}
    assert list(spicerack.authdns_active_hosts.hosts) == ["authdns1001.example.org", "authdns2001.example.org"]
    assert isinstance(spicerack.irc_logger, logging.Logger)
    assert isinstance(spicerack.sal_logger, logging.Logger)
    assert isinstance(spicerack.actions, ActionsDict)
    assert isinstance(spicerack.remote(), Remote)
    assert isinstance(spicerack.remote(installer=True), Remote)
    assert isinstance(spicerack.confctl("discovery"), ConftoolEntity)
    assert isinstance(spicerack.confctl("mwconfig"), ConftoolEntity)
    assert isinstance(spicerack.dns(), Dns)
    assert isinstance(spicerack.discovery("discovery-record"), Discovery)
    assert isinstance(spicerack.kubernetes("group", "cluster"), Kubernetes)
    assert isinstance(spicerack.mediawiki(), MediaWiki)
    assert isinstance(spicerack.mysql(), Mysql)
    assert isinstance(spicerack.mysql_legacy(), MysqlLegacy)
    assert isinstance(spicerack.redis_cluster("cluster"), RedisCluster)
    assert isinstance(
        spicerack.elasticsearch_clusters("search_eqiad", ("some_core_dc",)),
        ElasticsearchClusters,
    )
    assert isinstance(spicerack.admin_reason("Reason message", task_id="T12345"), Reason)
    assert isinstance(spicerack.puppet(mock.MagicMock(spec_set=RemoteHosts)), PuppetHosts)
    assert isinstance(
        spicerack.phabricator(get_fixture_path("phabricator", "valid.conf")),
        Phabricator,
    )
    assert isinstance(spicerack.prometheus(), Prometheus)
    assert isinstance(spicerack.thanos(), Thanos)
    assert isinstance(spicerack.debmonitor(), Debmonitor)
    assert isinstance(spicerack.ganeti(), Ganeti)
    assert isinstance(spicerack.requests_session("name"), Session)
    assert isinstance(
        spicerack.etcdctl(remote_host=mock.MagicMock(spec_set=RemoteHosts)),
        EtcdctlController,
    )
    assert isinstance(spicerack.kafka(), Kafka)
    service_catalog = spicerack.service_catalog()
    assert isinstance(service_catalog, Catalog)
    assert spicerack.service_catalog() is service_catalog  # Returned the cached instance
    assert isinstance(spicerack.apt_get(mock.MagicMock(spec_set=RemoteHosts)), AptGetHosts)
    assert isinstance(spicerack.lock(), NoLock)

    assert mocked_dns_resolver.Resolver.called

    with pytest.raises(AttributeError, match="AttributeError: 'Spicerack' object has no attribute 'nonexistent'"):
        # Test that non-existent accessors raise when there is no extender
        spicerack.nonexistent()


def test_spicerack_http_proxy():
    """An instance of Spicerack by default should not set any HTTP proxy."""
    spicerack = Spicerack(**SPICERACK_TEST_PARAMS)
    assert spicerack.http_proxy == ""
    assert spicerack.requests_proxies is None


@mock.patch("spicerack.gethostname", return_value="test.example.com")
@mock.patch("spicerack.Dns", autospec=True)
@mock.patch("spicerack.remote.Remote.query", autospec=True)
@mock.patch("spicerack.icinga.CommandFile", autospec=True)
def test_spicerack_icinga(mocked_command_file, mocked_remote_query, mocked_dns, mocked_hostname, monkeypatch):
    """An instance of Spicerack should allow to get an Icinga and IcingaHosts instances."""
    monkeypatch.setenv("SUDO_USER", "user1")
    icinga_server = mock.MagicMock(spec_set=RemoteHosts)
    icinga_server.hosts = "icinga-server.example.com"
    icinga_server.__len__.return_value = 1
    mocked_remote_query.return_value = icinga_server
    mocked_dns.return_value.resolve_cname.return_value = "icinga-server.example.com"
    mocked_command_file.return_value = "/var/lib/icinga/rw/icinga.cmd"

    spicerack = Spicerack(verbose=True, dry_run=False, **SPICERACK_TEST_PARAMS)

    assert spicerack.icinga_master_host.hosts == "icinga-server.example.com"
    assert isinstance(spicerack.icinga_hosts(["host1", "host2"]), IcingaHosts)
    mocked_hostname.assert_called_once_with()


@mock.patch("spicerack.get_puppet_ca_hostname", return_value="puppetmaster.example.com")
@mock.patch("spicerack.remote.Remote.query", autospec=True)
def test_spicerack_puppet_master(mocked_remote_query, mocked_get_puppet_ca_hostname):
    """An instance of Spicerack should allow to get a PuppetMaster instance."""
    host = mock.MagicMock(spec_set=RemoteHosts)
    host.__len__.return_value = 1
    mocked_remote_query.return_value = host
    spicerack = Spicerack(verbose=True, dry_run=False, **SPICERACK_TEST_PARAMS)

    assert isinstance(spicerack.puppet_master(), PuppetMaster)
    mocked_get_puppet_ca_hostname.assert_called_once_with()
    assert mocked_remote_query.called


@mock.patch("spicerack.get_ca_via_srv_record", return_value="puppetserver1001.example.org")
@mock.patch("spicerack.remote.Remote.query", autospec=True)
def test_spicerack_puppet_server(mocked_remote_query, mocked_get_ca_via_srv_record):
    """An instance of Spicerack should allow to get a PuppetServer instance."""
    host = mock.MagicMock(spec_set=RemoteHosts)
    host.__len__.return_value = 1
    mocked_remote_query.return_value = host
    spicerack = Spicerack(verbose=True, dry_run=False, **SPICERACK_TEST_PARAMS)

    assert isinstance(spicerack.puppet_server(), PuppetServer)
    assert mocked_remote_query.called
    assert mocked_get_ca_via_srv_record.called


@pytest.mark.parametrize(
    "manufacturer, manufacturer_class",
    (
        ("dell", RedfishDell),
        ("supermicro", RedfishSupermicro),
        ("fancy-but-not-supported-yet", None),
    ),
)
@mock.patch("spicerack.Netbox")
def test_spicerack_management_consoles(mocked_netbox, monkeypatch, manufacturer, manufacturer_class):
    """Should instantiate the instances that require the management password."""
    monkeypatch.setenv("MGMT_PASSWORD", "env_password")
    mocked_netbox.return_value.get_server.return_value.virtual = False
    mocked_netbox.return_value.api.ipam.ip_addresses.get.return_value = "10.0.0.1/16"
    mocked_netbox.return_value.get_server.return_value.as_dict.return_value = {
        "device_type": {"manufacturer": {"slug": manufacturer}}
    }

    spicerack = Spicerack(verbose=True, dry_run=False, **SPICERACK_TEST_PARAMS)

    assert spicerack.management_password == "env_password"
    assert isinstance(spicerack.ipmi("test-mgmt.example.com"), Ipmi)
    if manufacturer == "fancy-but-not-supported-yet":
        with pytest.raises(
            SpicerackError, match=f"The manufacturer {manufacturer} set in Netbox for test-mgmt01 is not supported."
        ):
            spicerack.redfish("test-mgmt01")
    else:
        assert isinstance(spicerack.redfish("test-mgmt01"), manufacturer_class)
        assert isinstance(spicerack.redfish("test-mgmt01", "root", "other_password"), manufacturer_class)
        assert mocked_netbox.called


@mock.patch("spicerack.Netbox")
def test_spicerack_redfish_not_physical(mocked_netbox, monkeypatch):
    """Should raise a SpicerackError if trying to get a management console for a non-physical device."""
    monkeypatch.setenv("MGMT_PASSWORD", "env_password")
    mocked_netbox.return_value.get_server.return_value.virtual = True
    spicerack = Spicerack(verbose=True, dry_run=False, **SPICERACK_TEST_PARAMS)

    with pytest.raises(SpicerackError, match="Host test-mgmt01 is not a Physical server, Redfish is not supported"):
        spicerack.redfish("test-mgmt01")

    assert mocked_netbox.called


def test_spicerack_management_password_cached(monkeypatch):
    """Should ask for the management_password only once and cache its result."""
    monkeypatch.setenv("MGMT_PASSWORD", "first_password")
    spicerack = Spicerack(verbose=True, dry_run=False, **SPICERACK_TEST_PARAMS)
    assert spicerack.management_password == "first_password"
    monkeypatch.setenv("MGMT_PASSWORD", "second_password")
    assert spicerack.management_password == "first_password"


@pytest.mark.parametrize(
    "read_write, token",
    (
        (False, "ro_token"),
        (True, "rw_token"),
    ),
)
@mock.patch("spicerack.Dns", autospec=True)
@mock.patch("spicerack.remote.Remote.query", autospec=True)
@mock.patch("pynetbox.api")
def test_spicerack_netbox(mocked_pynetbox, mocked_remote_query, mocked_dns, read_write, token):
    """Test instantiating Netbox abstraction."""
    netbox_server = mock.MagicMock(spec_set=RemoteHosts)
    netbox_server.hosts = "netbox-server.example.com"
    mocked_remote_query.return_value = netbox_server
    mocked_dns.return_value.resolve_cname.return_value = "netbox-server.example.com"
    mocked_pynetbox.reset_mock()

    spicerack = Spicerack(verbose=True, dry_run=False, **SPICERACK_TEST_PARAMS)

    assert isinstance(spicerack.netbox(read_write=read_write), Netbox)
    # Values from fixtures/netbox/config.yaml
    mocked_pynetbox.assert_called_once_with("https://netbox.example.com", token=token, threading=True)
    assert spicerack.netbox_master_host.hosts == "netbox-server.example.com"

    mocked_pynetbox.reset_mock()
    mocked_pynetbox.return_value.dcim.devices.get.return_value.device_role.slug = "server"
    assert isinstance(spicerack.netbox_server("host1"), NetboxServer)


@mock.patch("spicerack.remote.Remote.query", autospec=True)
def test_spicerack_dhcp_ok(mocked_remote_query):
    """It should return an instance of the DHCP class if created with the correct parameters."""
    mock_hosts = mock.MagicMock(spec_set=RemoteHosts)
    mock_hosts.__len__.return_value = 1
    mocked_remote_query.return_value = mock_hosts

    spicerack = Spicerack(verbose=True, dry_run=False, **SPICERACK_TEST_PARAMS)
    assert isinstance(spicerack.dhcp("codfw"), DHCP)
    assert mocked_remote_query.call_args.args[1] == "A:installserver and A:codfw"


@mock.patch("spicerack.remote.Remote.query", autospec=True)
def test_spicerack_dhcp_fallback(mocked_remote_query):
    """It should still create an instance of the DHCP class but fallback to eqiad if the query fails."""
    mock_hosts = mock.MagicMock(spec_set=RemoteHosts)
    mock_hosts.__len__.return_value = 1
    mocked_remote_query.side_effect = [RemoteError(), mock_hosts]

    spicerack = Spicerack(verbose=True, dry_run=False, **SPICERACK_TEST_PARAMS)
    assert isinstance(spicerack.dhcp("codfw"), DHCP)
    assert mocked_remote_query.call_args.args[1] == "A:installserver and A:eqiad"


def test_run_cookbook_no_callback():
    """It should raise a SpicerackError if there is no get_cookbook_callback defined."""
    spicerack = Spicerack(verbose=True, dry_run=False, **SPICERACK_TEST_PARAMS)
    with pytest.raises(SpicerackError, match="Unable to run other cookbooks, get_cookbook_callback is not set."):
        spicerack.run_cookbook("class_api.example", [])


@mock.patch("spicerack.Path.is_dir")
@mock.patch("spicerack.Repo")
def test_reposync(mocked_repo, mocked_is_dir):
    """Test spicerack.reposync."""
    spicerack = Spicerack(**SPICERACK_TEST_PARAMS)

    repo = mock.MagicMock(spec_set=Repo)
    repo.bare = False

    with pytest.raises(SpicerackError, match="Unknown repo missing_repo"):
        spicerack.reposync("missing_repo")

    mocked_is_dir.return_value = False
    with pytest.raises(SpicerackError, match=r"The repo directory \(/testrepo\) does not exist"):
        spicerack.reposync("testrepo")

    mocked_is_dir.return_value = True
    mocked_repo.return_value = repo
    with pytest.raises(SpicerackError, match=r"The repo directory \(/testrepo\) is not a bare git repository"):
        spicerack.reposync("testrepo")

    repo.bare = True
    assert isinstance(spicerack.reposync("testrepo"), RepoSync)


@mock.patch("spicerack.gethostname", return_value="test.example.com")
@mock.patch("spicerack.Dns", autospec=True)
@mock.patch("spicerack.remote.Remote.query", autospec=True)
@mock.patch("spicerack.icinga.CommandFile", autospec=True)
def test_spicerack_alerting(mocked_command_file, mocked_remote_query, mocked_dns, mocked_hostname, monkeypatch):
    """An instance of Spicerack should allow to get an AlertingHosts instance."""
    monkeypatch.setenv("SUDO_USER", "user1")
    icinga_server = mock.MagicMock(spec_set=RemoteHosts)
    icinga_server.hosts = "icinga-server.example.com"
    icinga_server.__len__.return_value = 1
    mocked_remote_query.return_value = icinga_server
    mocked_dns.return_value.resolve_cname.return_value = "icinga-server.example.com"
    mocked_command_file.return_value = "/var/lib/icinga/rw/icinga.cmd"

    spicerack = Spicerack(verbose=True, dry_run=False, **SPICERACK_TEST_PARAMS)

    assert spicerack.icinga_master_host.hosts == "icinga-server.example.com"
    assert isinstance(spicerack.alerting_hosts(["host1", "host2"]), AlertingHosts)
    mocked_hostname.assert_called_once_with()


@pytest.mark.parametrize("am_instance", (None, "production", "metricsinfra-eqiad1"))
def test_spicerack_alertmanager(am_instance):
    """An instance of Spicerack should allow to get an Alertmanager instance."""
    spicerack = Spicerack(
        verbose=True, dry_run=False, http_proxy="http://webproxy.eqiad.example:8080", **SPICERACK_TEST_PARAMS
    )

    config = load_yaml_config(spicerack.config_dir / "alertmanager" / "config.yaml")
    instance = config.get("instances").get(am_instance if am_instance else config.get("default_instance"))
    urls = instance.get("urls")

    http_authentication = None
    if "http_username" in instance and "http_password" in instance:
        http_authentication = HTTPBasicAuth(username=instance["http_username"], password=instance["http_password"])
    http_proxy = "http://webproxy.eqiad.example:8080" if instance.get("http_use_proxy") else None

    instance = spicerack.alertmanager(instance_name=am_instance)

    assert isinstance(instance, Alertmanager)
    assert instance._alertmanager_urls == urls  # pylint: disable=protected-access
    assert instance._http_session.auth == http_authentication  # pylint: disable=protected-access
    assert instance._http_session.proxies.get("http") == http_proxy  # pylint: disable=protected-access

    hosts_instance = spicerack.alertmanager_hosts(["host1", "host2"], instance_name=am_instance)
    assert isinstance(hosts_instance, AlertmanagerHosts)
    assert hosts_instance._alertmanager_urls == urls  # pylint: disable=protected-access
    assert hosts_instance._http_session.auth == http_authentication  # pylint: disable=protected-access
    assert hosts_instance._http_session.proxies.get("http") == http_proxy  # pylint: disable=protected-access


def test_spicerack_alertmanager_invalid_instance():
    """An instance of Spicerack should throw an error when given a nonexistent Alertmanager instance name."""
    spicerack = Spicerack(verbose=True, dry_run=False, **SPICERACK_TEST_PARAMS)
    with pytest.raises(SpicerackError):
        spicerack.alertmanager("nonexistent")


@pytest.mark.parametrize("am_instance", ("hosts-missing", "hosts-empty"))
def test_spicerack_alertmanager_instance_without_hosts(am_instance):
    """An instance of Spicerack should throw an error when given an Alertmanager instance with no hosts."""
    spicerack = Spicerack(verbose=True, dry_run=False, **SPICERACK_TEST_PARAMS)
    with pytest.raises(SpicerackError):
        spicerack.alertmanager(am_instance)


@mock.patch("spicerack.load_yaml_config")
def test_spicerack_alertmanager_no_default_instance(mocked_load_yaml_config):
    """An instance of Spicerack should throw an error when trying to access a non-configured default instance."""
    spicerack = Spicerack(verbose=True, dry_run=False, **SPICERACK_TEST_PARAMS)

    mocked_load_yaml_config.return_value = {"instances": {"nondefault": {"urls": ["https://alertmanager.example"]}}}

    with pytest.raises(SpicerackError):
        spicerack.alertmanager()


@pytest.mark.parametrize("ttl", (None, 3600))
@pytest.mark.parametrize("api_token_ro", ("", "sometoken"))
@pytest.mark.parametrize("cachedir", (None, ""))
@mock.patch("spicerack.load_yaml_config")
def test_spicerack_peeringdb(mocked_load_yaml_config, cachedir, api_token_ro, ttl, tmp_path):
    """An instance of Spicerack should allow to get a PeeringDB instance."""
    spicerack = Spicerack(verbose=True, dry_run=False, **SPICERACK_TEST_PARAMS)

    config = {}
    if api_token_ro:
        config["api_token_ro"] = api_token_ro
    if cachedir is not None:
        config["cachedir"] = str(tmp_path)
    mocked_load_yaml_config.return_value = config

    if ttl is not None:
        instance = spicerack.peeringdb(ttl=ttl)
    else:
        instance = spicerack.peeringdb()

    assert isinstance(instance, PeeringDB)


def test_spicerack_extender():
    """An instance of Spicerack with an extender should allow to access the extender accessors."""
    sys.path.append(str(get_fixture_path("external_modules")))
    loader_module = import_module("spicerack_extender")
    spicerack = Spicerack(extender_class=getattr(loader_module, "SpicerackExtender"), **SPICERACK_TEST_PARAMS)

    assert str(spicerack.cool_feature("Extender")) == "Extender is a cool feature!"

    with pytest.raises(AttributeError, match="'SpicerackExtender' object has no attribute 'nonexistent'"):
        # Test that non-existent accessors raise when there is an extender.
        spicerack.nonexistent()


def test_spicerack_lock(monkeypatch):
    """It should return an instance of spicerack.locking.Lock."""
    monkeypatch.setenv("USER", "")
    spicerack = Spicerack(etcd_config=get_fixture_path("locking", "config.yaml"), **SPICERACK_TEST_PARAMS)
    assert isinstance(spicerack.lock(), Lock)


def test_spicerack_private_lock():
    """It should return a lock instance for the spicerack modules and also cache it for re-use."""
    Spicerack.test_accessor = lambda self: getattr(self, "_spicerack_lock")
    spicerack = Spicerack(**SPICERACK_TEST_PARAMS)

    lock_1 = spicerack.test_accessor()
    lock_2 = spicerack.test_accessor()
    assert isinstance(lock_1, NoLock)
    assert lock_1 is lock_2  # Test that it returns the cached object



"""Class API get_runner raise cookbook."""
from spicerack.cookbook import CookbookBase


class GetRunnerRaise(CookbookBase):
    """Class API get_runner raise cookbook."""

    def get_runner(self, args):
        """As required by the parent class."""
        raise RuntimeError("get_runner raise")



"""Module to manage distributed locks."""
from __future__ import annotations

import json
import logging
import os
import uuid
from contextlib import contextmanager
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from pathlib import Path
from typing import Any, Iterator, Optional, Union

import etcd
from wmflib.config import load_yaml_config

from spicerack.decorators import retry
from spicerack.exceptions import SpicerackCheckError, SpicerackError

logger = logging.getLogger(__name__)
KEYS_BASE_PATH: str = "/spicerack/locks"
"""The base path for the lock keys."""
COOKBOOKS_PREFIX: str = "cookbooks"
"""The path prefix to use for the key of locks acquired by spicerack for each cookbook execution."""
COOKBOOKS_CUSTOM_PREFIX: str = "custom"
"""The path prefix to use for the key of locks acquired from inside a cookbook."""
SPICERACK_PREFIX: str = "modules"
"""The path prefix to use for the key of locks acquired from inside spicerack modules."""
ALLOWED_PREFIXES: tuple[str, ...] = (COOKBOOKS_PREFIX, COOKBOOKS_CUSTOM_PREFIX, SPICERACK_PREFIX)
"""The allowed values for the prefix parameter to be used as path prefix for the locks keys."""
ETCD_WRITER_LOCK_KEY: str = "etcd"
"""The path prefix of the short-term keys to acquire an exclusive lock to write to etcd."""


def get_lock_instance(
    *, config_file: Optional[Path], prefix: str, owner: str, dry_run: bool = True
) -> Union[Lock, NoLock]:
    """Get a lock instance based on the configuration file and prefix.

    Arguments:
        config_file: the path to the configuration file for the locking backend or :py:data:`None` to disable the
            locking support and return a :py:class:`spicerack.locking.NoLock` instance. When the configuration file is
            present a :py:class:`spicerack.locking.Lock` instance is returned instead. The configuration is also
            automatically merged with the ``~/.etcdrc`` config file of the running user, if present.
        prefix: the name of the directory to use to prefix the lock. Must be one of
            :py:const:`spicerack.locking.ALLOWED_PREFIXES`.
        owner: a way to identify the owner of the lock, usually in the form ``{user}@{hostname} [{pid}]``.
        dry_run: whether this is a DRY-RUN.

    Raises:
        spicerack.locking.LockError: if the provided prefix is not one of the allowed ones.

    Returns:
        The locking instance or a dummy instance that has the same API of the locking one but does nothing if the
        config_file.

    """
    if prefix not in ALLOWED_PREFIXES:
        raise LockError(f"Invalid prefix {prefix}, must be one of: {ALLOWED_PREFIXES}")

    if config_file is not None:
        user = os.environ.get("USER", "")
        config = load_yaml_config(config_file)
        config.update(load_yaml_config(Path(f"~{user}/.etcdrc").expanduser(), raises=False))
        return Lock(prefix=prefix, config=config, owner=owner, dry_run=dry_run)

    return NoLock()


class LockError(SpicerackError):
    """Generic exception for errors of this module."""


class InvalidLockError(LockError):
    """Exception raised when the parameters for a lock are invalid."""


class LockUnreadableError(LockError):
    """Exception raised when unable to properly parse an existing lock from the backend."""


class LockUnwritableError(LockError):
    """Exception raised when unable to properly serialize the lock in order to save it in the backend."""


class LockExistingError(LockError):
    """Exception raised if a lock with the same ID already exists for the given key."""


class EtcdLockUnavailableError(SpicerackCheckError):
    """Exception raised when unable to acquire the etcd lock, the operation should be retried."""


class LockUnavailableError(SpicerackCheckError):
    """Exception raised when unable to acquire the Spicerack lock, the operation should be retried."""


class Lock:
    """Manage a Spicerack lock.

    Example object created in the backend (etcd)::

        '/spicerack/locks/cookbooks/sre.foo.bar' => {
            '4e2677c7-541a-4f9e-afbb-cfdb69c440a1': {
                'created': '2023-07-16 16:46:52.161053',
                'owner': 'user@host [12345]',
                'concurrency': 5,
                'ttl': 120,
            }
        }

    """

    def __init__(self, *, config: dict, prefix: str, owner: str, dry_run: bool = True) -> None:
        """Initialize the instance.

        Arguments:
            config: the etcd configuration dictionary to be passed to the etcd client.
            prefix: the name of the directory to use to prefix the lock. Must be one of
                :py:const:`spicerack.locking.ALLOWED_PREFIXES`.
            owner: a way to identify the owner of the lock, usually in the form ``{user}@{hostname} [{pid}]``.
            dry_run: whether this is a DRY-RUN.

        Raises:
            spicerack.locking.LockError: if the provided prefix is not one of the allowed ones.

        """
        if prefix not in ALLOWED_PREFIXES:
            raise LockError(f"Invalid prefix {prefix}, must be one of: {ALLOWED_PREFIXES}")

        config["lock_prefix"] = KEYS_BASE_PATH  # Force the lock prefix
        self._etcd = etcd.Client(**config)
        self._owner = owner
        self._prefix = prefix
        self._dry_run = dry_run

    def get(self, name: str) -> KeyLocks:
        """Get the existing locks for the given name. If missing returns a new object with no locks.

        Arguments:
            name: the lock name, cannot contain ``/`` as that's a directory separator in the data structure.

        Raises:
            spicerack.locking.LockError: for any etcd errors beside the key not found one.

        Returns:
            the existing locks for the given name.

        """
        key = self._get_key(name)
        try:
            return KeyLocks.from_json(
                key, self._etcd.read(key, timeout=self._etcd.read_timeout).value  # pylint: disable=no-member
            )
        except (KeyError, etcd.EtcdKeyNotFound):  # Does not exist, create a new one without any locks
            return KeyLocks(key=key)
        except etcd.EtcdException as e:
            raise LockError(f"Failed to get key {key}") from e

    @contextmanager
    def acquired(self, name: str, *, concurrency: int, ttl: int) -> Iterator[None]:
        """Context manager to perform actions while holding a lock for the given name and parameters.

        Examples::

            # Get a lock for a given name allowing to have other 2 concurrent runs in parallel with a TTL of 30 minutes
            with lock.acquired("some.lock.identifier", concurrency=3, ttl=1800):
                # Do something

            # Acquire an exclusive lock for the given name with a TTL of 1 hour
            with lock.acquired("some.lock.identifier", concurrency=1, ttl=1800):
                # Do something

        Arguments:
            name: the lock name, cannot contain ``/`` as that's a directory separator in the data structure.
            concurrency: how many concurrent clients can hold the same lock. If set to zero it means that there is no
                concurrency limit and an infinite number of clients can run concurrently, but they are tracked.
            ttl: the amount of seconds this lock is valid for. All locks that have passed their TTL are considered
                expired and will be automatically removed.

        Yields:
            Nothing, it just gives back control to the client's code while holding the lock.

        """
        lock_id = self.acquire(name, concurrency=concurrency, ttl=ttl)
        try:
            yield
        finally:
            self.release(name, lock_id)

    def acquire(self, name: str, *, concurrency: int, ttl: int) -> str:
        """Acquire a lock for the given name and parameters and return the lock identifier.

        Examples::

            # Get a lock for a given name allowing to have other 2 concurrent runs in parallel with a TTL of 30 minutes
            lock_id = lock.acquire("some.lock.identifier", concurrency=3, ttl=1800)
            # Do something
            lock.release("some.lock.identifier", lock_id)

            # Acquire an exclusive lock for the given name with a TTL of 1 hour
            lock_id = lock.acquire("some.lock.identifier", concurrency=1, ttl=1800)
            # Do something
            lock.release("some.lock.identifier", lock_id)

        Arguments:
            name: the lock name, cannot contain ``/`` as that's a directory separator in the data structure.
            concurrency: how many concurrent clients can hold the same lock. If set to zero it means that there is no
                concurrency limit and an infinite number of clients can run concurrently, but they are tracked.
            ttl: the amount of seconds this lock is valid for. All locks that have passed their TTL are considered
                expired and will be automatically removed.

        Returns:
            the lock unique identifier.

        """
        lock = ConcurrentLock(concurrency=concurrency, owner=self._owner, ttl=ttl)
        logger.debug("Acquiring lock for key %s: %s", name, lock)
        self._acquire_lock(name, lock)

        return lock.uuid

    def release(self, name: str, lock_id: str) -> None:
        """Release the lock identified by the lock ID, best effort. The lock will expire anyway.

        See the documentation for :py:meth:`spicerack.locking.Lock.acquire` for usage examples.

        Arguments:
            name: the lock name, cannot contain ``/`` as that's a directory separator in the data structure.
            lock_id: the ID identifying the lock.

        """
        logger.debug("Releasing lock for key %s with ID %s", name, lock_id)
        try:
            with self._etcd_locked():
                key_lock = self.get(name)
                lock = key_lock.remove(lock_id)
                if lock is not None:
                    self._set(key_lock)
                    logger.info("Released lock for key %s: %s", key_lock.key, lock)

        except Exception as e:  # pylint: disable=broad-exception-caught
            logger.error("Failed to release lock for key %s and ID %s: %s", self._get_key(name), lock_id, e)

    @contextmanager
    def _etcd_locked(self) -> Iterator[None]:
        """Perform an action while holding an exclusive write lock to etcd for operations in this module.

        Yields:
            Nothing, it just gives back control to the client's code.

        Raises:
            spicerack.locking.EtcdLockUnavailableError: if unable to acquire the etcd lock.

        """
        if self._dry_run:
            yield
        else:
            etcd_lock = self._acquire_etcd_lock()
            try:
                yield
            finally:
                etcd_lock.release()

    def _get_key(self, name: str) -> str:
        """Return the key to be used for the given lock name.

        Arguments:
            name: the lock name, cannot contain ``/`` as that's a directory separator in the data structure.

        Returns:
            the lock key.

        """
        if not name:
            raise InvalidLockError("The lock name cannot be empty")

        if "/" in name:
            raise InvalidLockError(f"The lock name cannot contain '/', got: {name}")

        if self._prefix == SPICERACK_PREFIX:  # Ensure the key represent a spicerack module
            if not name.startswith("spicerack."):
                raise InvalidLockError(
                    f"Locks with prefix '{SPICERACK_PREFIX}' must have names starting with 'spicerack.', got: {name}"
                )

        return "/".join([KEYS_BASE_PATH, self._prefix, name])

    def _set(self, lock: KeyLocks) -> None:
        """Set the locks for the given key, deletes the key if there are no more locks.

        Arguments:
            lock: the lock instance that represent the locks for a given key.

        """
        if self._dry_run:
            logger.info("Skipping lock acquire/release in DRY-RUN mode")
        else:
            if lock.locks:
                self._etcd.set(lock.key, lock.to_json())
            else:  # No locks present, delete the key to keep etcd clean
                self._etcd.delete(lock.key)

    @retry(  # Retry for 2 minutes
        tries=15,
        delay=timedelta(seconds=1),
        backoff_mode="linear",
        exceptions=(EtcdLockUnavailableError,),
        failure_message="Unable to acquire etcd lock",
    )
    def _acquire_etcd_lock(self) -> etcd.Lock:
        """Acquire the etcd lock to write the Spicerack lock objects.

        Returns:
            the lock object so it can get released after the Spicerack object has been saved.

        Raises:
            spicerack.locking.EtcdLockUnavailableError: if unable to acquire the etcd lock.

        """
        try:
            etcd_lock = etcd.Lock(self._etcd, ETCD_WRITER_LOCK_KEY)
            etcd_lock.acquire(blocking=False, lock_ttl=15)  # If blocking=True blocks forever even if setting timeout=N
        except etcd.EtcdException as e:
            try:  # Best effort to release the object in the lock queue
                etcd_lock.release()
            except etcd.EtcdException as release_exception:
                logger.warning(
                    "Failed to release etcd attempted lock queued in %s with value %s: %s",
                    etcd_lock.path,
                    etcd_lock.uuid,
                    release_exception,
                )

            raise EtcdLockUnavailableError(f"{e}") from e

        if not etcd_lock.is_acquired:
            try:  # Best effort to release the object in the lock queue
                etcd_lock.release()
            except etcd.EtcdException as e:
                logger.warning(
                    "Failed to release etcd lock queued in %s with value %s: %s", etcd_lock.path, etcd_lock.uuid, e
                )

            raise EtcdLockUnavailableError("Lock already taken")

        return etcd_lock

    @retry(  # Retry for more or less half an hour
        tries=27,
        delay=timedelta(seconds=5),
        backoff_mode="linear",
        exceptions=(LockUnavailableError,),
        failure_message="Unable to acquire lock",
    )
    def _acquire_lock(self, name: str, lock: ConcurrentLock) -> None:
        """Try to acquire the spicerack lock in etcd, retrying on failure for some time.

        Arguments:
            name: the lock name, cannot contain ``/`` as that's a directory separator in the data structure.
            lock: the lock to be acquired.

        Raises:
            spicerack.locking.LockUnavailableError: if unable to acquire the etcd lock to write the actual lock.

        """
        with self._etcd_locked():
            key_lock = self.get(name)
            key_lock.add(lock)
            self._set(key_lock)
            logger.info("Acquired lock for key %s: %s", key_lock.key, lock)


class NoLock:
    """A noop locking class that does nothing.

    Has the same APIs of the :py:class:`spicerack.locking.Lock`, to be used when locking support is disabled.

    """

    def __init__(self, *_args: Any, **_kwargs: Any) -> None:
        """Initialize the instance.

        Arguments:
            *_args: accept any positional arguments that :py:meth:`spicerack.locking.Lock.__init__` accepts.
            **_kwargs: accept any keyword argument that :py:meth:`spicerack.locking.Lock.__init__` accepts.

        """

    def get(self, name: str, *_args: Any, **_kwags: Any) -> KeyLocks:
        """Dummy method that just returns an empty KeyLocks.

        Arguments:
            name: the lock name, cannot contain ``/`` as that's a directory separator in the data structure.
            *_args: accept any positional arguments that :py:meth:`spicerack.locking.Lock.acquire` accepts.
            **_kwargs: accept any keyword argument that :py:meth:`spicerack.locking.Lock.acquire` accepts.

        Returns:
            an empty instance.

        """
        return KeyLocks(key=name)

    @contextmanager
    def acquired(self, *_args: Any, **_kwargs: Any) -> Iterator[None]:
        """Context manager that just yields.

        Arguments:
            *_args: accept any positional arguments that :py:meth:`spicerack.locking.Lock.acquired` accepts.
            **_kwargs: accept any keyword argument that :py:meth:`spicerack.locking.Lock.acquired` accepts.

        Yields:
            Nothing, it just gives back control to the client's code.

        """
        yield

    def acquire(self, *_args: Any, **_kwags: Any) -> str:
        """Dummy method that just returns an empty string.

        Arguments:
            *_args: accept any positional arguments that :py:meth:`spicerack.locking.Lock.acquire` accepts.
            **_kwargs: accept any keyword argument that :py:meth:`spicerack.locking.Lock.acquire` accepts.

        Returns:
            an empty string.

        """
        return ""

    def release(self, *_args: Any, **_kwags: Any) -> None:
        """Dummy method that does nothing.

        Arguments:
            *_args: accept any positional arguments that :py:meth:`spicerack.locking.Lock.release` accepts.
            **_kwargs: accept any keyword argument that :py:meth:`spicerack.locking.Lock.release` accepts.

        """


@dataclass
class KeyLocks:
    """Manage a lock object for a given key with concurrent locks.

    Arguments:
        key: the lock full key path to be used in the backend.
        locks: the concurrent locks for the given key.

    """

    key: str
    locks: dict[str, ConcurrentLock] = field(default_factory=dict)

    @classmethod
    def from_json(cls, key: str, json_str: str) -> KeyLocks:
        """Create an instance of this class from a JSON string.

        Arguments:
            key: the lock full key path to be used in the backend.
            json_str: the JSON serialization of the lock object.

        Returns:
            the lock instance.

        Raises:
            spicerack.locking.LockUnreadableError: if unable to decode the JSON lock.

        """
        try:
            json_obj = json.loads(json_str)
        except json.JSONDecodeError as e:
            raise LockUnreadableError(f"Unable to JSON-deserialize the lock for key {key}, got:\n{json_str}") from e

        locks = {lock_id: ConcurrentLock.from_dict(lock_id, lock_obj) for lock_id, lock_obj in json_obj.items()}
        return cls(key, locks)

    def to_json(self) -> str:
        """Return the JSON serialization of the current instance.

        Returns:
            a string with a JSON-serialized object.

        Raises:
            spicerack.locking.LockUnwritableError: if unable to convert the curent instance to JSON.

        """
        locks = {lock.uuid: lock.to_dict() for lock in self.locks.values()}
        try:
            return json.dumps(locks)
        except TypeError as e:
            raise LockUnwritableError("Unable to JSON-serialize the current instance") from e

    def add(self, lock: ConcurrentLock) -> None:
        """Add the concurrent lock to the lock object if all criteria are met.

        Arguments:
            lock: the concurrent lock to add.

        Raises:
            spicerack.locking.LockUnavailableError: when the max concurrency has been reached and is not possible to
                acquire a lock.
            spicerack.locking.LockExistingError: if a lock with the same ID already exists.

        """
        expired_uuids = []
        min_concurrency_lock = None
        for other_lock in self.locks.values():
            if other_lock.created + timedelta(seconds=other_lock.ttl) < datetime.utcnow():
                expired_uuids.append(other_lock.uuid)
                continue

            if (
                other_lock.concurrency != 0
                and len(self.locks) >= other_lock.concurrency
                and (min_concurrency_lock is None or other_lock.concurrency < min_concurrency_lock.concurrency)
            ):
                min_concurrency_lock = other_lock

        if min_concurrency_lock is not None:
            existing_locks = "\n".join(f"    {lock_obj.created}: {lock_obj.owner}" for lock_obj in self.locks.values())
            raise LockUnavailableError(
                f"{lock} for key {self.key}.\nThere are already {len(self.locks)} concurrent locks and the "
                f"concurrency allowed is {min_concurrency_lock.concurrency}:\n{existing_locks}"
            )

        for expired_uuid in expired_uuids:
            expired_lock = self.locks.pop(expired_uuid)
            logger.info("Releasing expired lock for key %s: %s", self.key, expired_lock)

        if lock.uuid in self.locks:
            raise LockExistingError(
                f"{lock} for key {self.key}. A lock with the same ID {lock.uuid} already exists: "
                f"{self.locks[lock.uuid]}"
            )

        if lock.concurrency != 0 and len(self.locks) >= lock.concurrency:
            raise LockUnavailableError(
                f"{lock} for key {self.key}. There are already {len(self.locks)} concurrent locks and a concurrency "
                f"of {lock.concurrency} was requested."
            )

        lock.update_created()
        self.locks[lock.uuid] = lock

    def remove(self, lock_id: str) -> Optional[ConcurrentLock]:
        """Remove the concurrent lock identified by the lock ID.

        Arguments:
            lock_id: the ID identifying the lock.

        Returns:
            the removed concurrent lock if the lock was found and removed from the instance, it means that the backend
            needs to be updated.
            :py:data:`None` if the concurrent lock was not found and there is no need to update the backend.

        """
        if lock_id in self.locks:
            return self.locks.pop(lock_id)

        logger.warning("Lock for key %s and ID %s not found. Unable to release it. Was expired?", self.key, lock_id)
        return None


@dataclass
class ConcurrentLock:
    """A single concurrent lock object.

    Arguments:
        concurrency: how many concurrent locks can be acquired for the same key. If set to zero it means that there is
            no concurrency limit and an infinite number of clients can run concurrently, but they are tracked.
        ttl: the time to live in seconds of the lock. If expired it will be automatically discarded.
        owner: a way to identify the owner of the lock, usually in the form ``{user}@{hostname} [{pid}]``.
        uuid: the unique identifier of the concurrent lock.
        created: when the lock has been created, used for checking if the TTL has expired.

    """

    concurrency: int
    owner: str
    ttl: int
    uuid: str = field(default_factory=lambda: str(uuid.uuid4()))
    created: datetime = field(default_factory=datetime.utcnow)

    def __post_init__(self) -> None:
        """According to Python's dataclass API to validate the arguments.

        See Also:
            https://docs.python.org/3/library/dataclasses.html#post-init-processing

        """
        if self.concurrency < 0:
            raise InvalidLockError(
                "Unable to create concurrent lock, argument 'concurrency' must be a non-negative integer, got "
                f"{self.concurrency}."
            )
        if self.ttl <= 0:
            raise InvalidLockError(
                f"Unable to create concurrent lock, argument 'ttl' must be a positive integer, got {self.ttl}."
            )
        if self.created > datetime.utcnow():
            raise InvalidLockError(
                f"Unable to create concurrent lock, argument 'created' cannot be in the future, got {self.created}."
            )

    def __str__(self) -> str:
        """String representation of the concurrent lock.

        Returns:
            the string representation.

        """
        return str(self.to_dict())

    def update_created(self) -> None:
        """Update the created time to now."""
        self.created = datetime.utcnow()

    def to_dict(self) -> dict[str, Union[str, int]]:
        """Returns the dict representation of the object suitable for JSON serialization.

        Intentionally not using ``dataclasses.asdict()`` to skip self.uuid and to convert the created datetime to
        string.

        Returns:
            the object as dict with the created datetime converted to string.

        """
        return {"concurrency": self.concurrency, "created": str(self.created), "owner": self.owner, "ttl": self.ttl}

    @classmethod
    def from_dict(cls, lock_id: str, lock_obj: dict[str, Union[str, int]]) -> ConcurrentLock:
        """Get a ConcurrentLock instance from a dict, suitable to be used to JSON deserialize.

        Arguments:
            lock_id: the concurrent lock unique identifier.
            lock_obj: the lock dictionary object.

        Returns:
            a concurrent lock instance.

        Raises:
            spicerack.locking.InvalidLockError: if unable to parse the created datetime.

        """
        params: dict = {**lock_obj}
        params["uuid"] = lock_id
        try:
            params["created"] = datetime.fromisoformat(str(lock_obj["created"]))
        except ValueError as e:
            raise InvalidLockError(
                "Unable to create concurent lock from dict, key 'created' cannot be parsed as a datetime object, "
                f"got {lock_obj['created']}."
            ) from e

        return cls(**params)



"""Confctl module tests."""
from unittest import mock

import pytest
from conftool.tests.unit import MockBackend

from spicerack import confctl
from spicerack.tests import get_fixture_path


class TestConfctl:
    """Confctl test class."""

    def setup_method(self):
        """Setup a Confctl instance with a mocked conftool backend and driver."""
        # pylint: disable=attribute-defined-outside-init
        self.conftool_backend = MockBackend({})
        confctl.kvobject.KVObject.backend = self.conftool_backend
        confctl.kvobject.KVObject.config = confctl.configuration.Config(driver="")
        config = get_fixture_path("confctl", "config.yaml")
        schema = get_fixture_path("confctl", "schema.yaml")
        with mock.patch("spicerack.confctl.kvobject.KVObject.setup"):
            self.confctl = confctl.Confctl(config=config, schema=schema, dry_run=False)
            self.entity = self.confctl._schema.entities["discovery"]  # pylint: disable=protected-access

        self.entity.query = mock.MagicMock(return_value=[self.entity("test", "dnsdisc")])
        self.discovery = confctl.ConftoolEntity(self.entity, dry_run=False)
        self.discovery_dry_run = confctl.ConftoolEntity(self.entity)

    @pytest.fixture
    def generated_entities(self, request):
        """Fixture that generates entities to use in tests."""
        entities = []
        for name, attrs in request.param.items():
            obj = self.entity("test", name)
            for k, v in attrs.items():
                setattr(obj, k, v)
            entities.append(obj)
        return entities

    def test_get_existing(self):
        """Calling get() should return the object matched by the tags."""
        for obj in self.discovery.get(dnsdisc="test"):
            assert obj.tags == {"dnsdisc": "test"}

    def test_get_non_existing(self):
        """Calling get() without matches should raise ConfctlError."""
        self.entity.query.return_value = []
        with pytest.raises(confctl.ConfctlError, match="No match found"):
            list(self.discovery.get(dnsdisc="non-existing"))

    def test_update_ok(self):
        """Calling update() should update the objects matched by the tags."""
        self.discovery.update({"pooled": True}, dnsdisc="test")
        assert list(self.discovery.get(dnsdisc="test"))[0].pooled

    def test_update_dry_run(self):
        """Calling update() in dry_run mode should not update the objects matched by the tags."""
        list(self.discovery_dry_run.get(dnsdisc="test"))[0].update = mock.MagicMock(side_effect=Exception("test"))
        self.discovery_dry_run.update({"pooled": True}, dnsdisc="test")

    @pytest.mark.parametrize(
        "exc_class, message",
        (
            (confctl.BackendError, "Error writing to etcd"),
            (Exception, "Generic error in conftool"),
        ),
    )
    def test_update_errors(self, exc_class, message):
        """Calling update() should raise ConfctlError if there is an error in the backend."""
        list(self.discovery.get(dnsdisc="test"))[0].update = mock.MagicMock(side_effect=exc_class("test"))

        with pytest.raises(confctl.ConfctlError, match=message):
            self.discovery.update({"pooled": True}, dnsdisc="test")

    def test_set_and_verify_ok(self):
        """It should update the objects matched by the tags and check them."""
        self.discovery.set_and_verify("pooled", True, dnsdisc="test")
        assert list(self.discovery.get(dnsdisc="test"))[0].pooled

    def test_set_and_verify_fail(self):
        """It should raise ConfctlError if failing to check the modified objects."""
        list(self.discovery.get(dnsdisc="test"))[0].update = mock.MagicMock()  # Don't allow to update the record
        with pytest.raises(
            confctl.ConfctlError,
            match="Conftool key pooled has value 'False', expecting 'True'",
        ):
            self.discovery.set_and_verify("pooled", True, dnsdisc="test")

    def test_set_and_verify_dry_run(self):
        """In dry_run mode it should not update the objects and not raise on failure to check them."""
        self.discovery_dry_run.set_and_verify("pooled", True, dnsdisc="test")
        assert not list(self.discovery.get(dnsdisc="test"))[0].pooled

    @pytest.mark.parametrize(
        "generated_entities",
        [{"foo": {"pooled": True}, "bar": {"pooled": False}}],
        indirect=True,
    )
    def test_filter_objects_ok(self, generated_entities):
        """It should correctly filter values based on their status."""
        self.entity.query = mock.MagicMock(return_value=generated_entities)
        to_select = [e for e in generated_entities if e.pooled is True]
        assert list(self.discovery.filter_objects({"pooled": True}, dnsdisc="test")) == to_select

    @pytest.mark.parametrize(
        "generated_entities",
        [{"foo": {"pooled": True}, "bar": {"pooled": True}}],
        indirect=True,
    )
    def test_filter_objects_empty(self, generated_entities):
        """It should return an empty generator if no object matches the request."""
        self.discovery._select = mock.MagicMock(return_value=generated_entities)  # pylint: disable=protected-access
        # pylint: disable=use-implicit-booleaness-not-comparison
        assert list(self.discovery.filter_objects({"pooled": False}, dnsdisc="test")) == []

    def test_filter_objects_bad_tags(self):
        """It should raise ConfctlError if no object is found in the datastore."""
        self.entity.query.return_value = []
        with pytest.raises(confctl.ConfctlError, match="No match found"):
            list(self.discovery.filter_objects({"pooled": True}, dnsdisc="test"))

    def test_filter_objects_fail(self):
        """Should raise an exception if the conftool query fails."""
        # Errors in the query get propagated
        self.entity.query = mock.MagicMock(side_effect=ValueError("something"))
        with pytest.raises(ValueError, match="something"):
            list(self.discovery.filter_objects({"pooled": True}, dnsdisc="test"))

    def test_filter_objects_bad_field(self):
        """Should raise an exception if a non-existent field is asked."""
        with pytest.raises(confctl.ConfctlError, match='Could not find property "pinkunicorn"'):
            list(self.discovery.filter_objects({"pinkunicorn": True}, dnsdisc="test"))

    @pytest.mark.parametrize(
        "generated_entities",
        [{"foo": {"pooled": True}, "bar": {"pooled": True}}],
        indirect=True,
    )
    def test_update_objects_ok(self, generated_entities):
        """Test a successful update of a list of objects."""
        self.discovery.update_objects({"pooled": False}, generated_entities)
        for obj in generated_entities:
            assert obj.pooled is False

    @pytest.mark.parametrize(
        "generated_entities",
        [{"foo": {"pooled": True}, "bar": {"pooled": True}}],
        indirect=True,
    )
    @pytest.mark.parametrize(
        "exc_class, message",
        (
            (confctl.BackendError, "Error writing to etcd"),
            (Exception, "Generic error in conftool"),
        ),
    )
    def test_update_objects_fail(self, generated_entities, exc_class, message):
        """An error in the backend should raise an exception."""
        generated_entities[1].update = mock.MagicMock(side_effect=exc_class(message))
        with pytest.raises(confctl.ConfctlError):
            self.discovery.update_objects({"pooled": False}, generated_entities)
        # Warning: the first object was updated
        assert generated_entities[0].pooled is False

    @pytest.mark.parametrize(
        "generated_entities",
        [{"foo": {"pooled": True}, "bar": {"pooled": True}}],
        indirect=True,
    )
    def test_update_objects_dry_run(self, generated_entities):
        """Under dry run, no update happens."""
        self.discovery_dry_run.update_objects({"pooled": False}, generated_entities)
        assert generated_entities[0].pooled is True

    @pytest.mark.parametrize(
        "generated_entities",
        [{"foo": {"pooled": True}, "bar": {"pooled": False}}],
        indirect=True,
    )
    def test_change_and_revert(self, generated_entities):
        """Test the contextmanager."""
        self.entity.query.return_value = generated_entities
        with self.discovery.change_and_revert("pooled", True, False, name="foo|bar") as pooled:
            # Check that only foo is present in the returned list
            assert len(pooled) == 1
            assert pooled[0].name == "foo"
            assert pooled[0].pooled is False
        assert pooled[0].pooled is True



"""MySQL shell module."""

import logging
from collections.abc import Iterator
from dataclasses import dataclass
from datetime import datetime, timedelta
from textwrap import dedent
from typing import Any, Optional

from ClusterShell.MsgTree import MsgTreeElem
from cumin import NodeSet
from wmflib.constants import CORE_DATACENTERS
from wmflib.interactive import ask_confirmation

from spicerack.decorators import retry
from spicerack.exceptions import SpicerackError
from spicerack.remote import Remote, RemoteExecutionError, RemoteHosts, RemoteHostsAdapter

REPLICATION_ROLES: tuple[str, ...] = ("master", "slave", "standalone")
"""Valid replication roles."""
CORE_SECTIONS: tuple[str, ...] = (
    "s1",
    "s2",
    "s3",
    "s4",
    "s5",
    "s6",
    "s7",
    "s8",
    "x1",
    "es6",
    "es7",
)
"""Valid MySQL RW core sections (external storage RO, parser cache, x2 and misc sections are not included here)."""

logger = logging.getLogger(__name__)


class MysqlLegacyError(SpicerackError):
    """Custom exception class for errors of this module."""


class MysqlLegacyReplagError(MysqlLegacyError):
    """Custom exception class for errors related to replag in this module."""


@dataclass
class ReplicationInfo:
    """Represent the minimum replication information needed to restore a replication from a given point.

    Arguments:
        primary: the FQDN of the primary host from where to replicate from.
        binlog: the binlog file to replicate from.
        position: the binlog position to replicate from.
        port: the port of the master from where to replicate from.

    """

    primary: str
    binlog: str
    position: int
    port: int


class Instance:
    """Class to manage MariaDB single intances and multiinstances."""

    def __init__(self, host: RemoteHosts, *, name: str = "") -> None:
        """Initialize the instance.

        Arguments:
            host: the RemoteHosts instance that contains this MariaDB SingleInstance.
            name: the name of the instance in a multiinstance context. Leave it empty for single instances.

        """
        if len(host) > 1:
            raise NotImplementedError("Only single hosts are currently supported. Got {len(host)}.")

        self.host = host
        self.name = name
        self._primary = ""
        self._mysql = "/usr/local/bin/mysql"

        if self.name:
            self._sock = f"/run/mysqld/mysqld.{self.name}.sock"
            self._service = f"mariadb@{self.name}.service"
            self._data_dir = f"/srv/sqldata.{self.name}"
        else:
            self._sock = "/run/mysqld/mysqld.sock"
            self._service = "mariadb.service"
            self._data_dir = "/srv/sqldata"

    def run_query(self, query: str, database: str = "", **kwargs: Any) -> Any:
        """Execute the query via Remote.

        Arguments:
            query: the mysql query to be executed. Double quotes must be already escaped.
            database: the optional database to use for the query execution.
            **kwargs: any additional argument is passed to :py:meth:`spicerack.remote.RemoteHosts.run_sync`. By default
                the ``print_progress_bars`` and ``print_output`` arguments are set to :py:data:`False`.

        Returns:
            The result of the remote command execution.

        Raises:
            spicerack.remote.RemoteExecutionError: if the query execution via SSH returns a non-zero exit code.

        """
        command = f'{self._mysql} --socket {self._sock} --batch --execute "{query}" {database}'.strip()
        kwargs.setdefault("print_progress_bars", False)
        kwargs.setdefault("print_output", False)
        try:
            return self.host.run_sync(command, **kwargs)
        except RemoteExecutionError as e:
            raise MysqlLegacyError(f"Failed to run '{query}' on {self.host}") from e

    def stop_slave(self) -> None:
        """Stops mariadb replication."""
        self.run_query("STOP SLAVE")

    def start_slave(self) -> None:
        """Starts mariadb replication."""
        self.run_query("START SLAVE")

    def show_slave_status(self) -> dict:
        """Returns the output of show slave status formatted as a dict.

        Returns:
            the current slave status for the instance.

        """
        result = {}
        rows: int = 0
        sql = r"SHOW SLAVE STATUS\G"
        response = list(self.run_query(sql, is_safe=True))
        if not response:
            raise MysqlLegacyError("SHOW SLAVE STATUS seems to have been executed on a master.")

        for line in response[0][1].message().decode("utf8").splitlines():
            if "***************************" in line:
                rows += 1
                continue
            if rows > 1:
                raise NotImplementedError("Multisource setup are not implemented.")

            key, value = line.split(":", 2)
            result[key.strip()] = value.strip()
        return result

    def use_gtid(self) -> None:
        """Runs SQL to use GTID."""
        self.run_query("CHANGE MASTER TO MASTER_USE_GTID=Slave_pos")

    def stop_mysql(self) -> Iterator[tuple[NodeSet, MsgTreeElem]]:
        """Stops mariadb service.

        Returns:
            The results of the remote status command.

        """
        return self.host.run_sync(f"/usr/bin/systemctl stop {self._service}")

    def status_mysql(self) -> Iterator[tuple[NodeSet, MsgTreeElem]]:
        """Stops mariadb service.

        Returns:
            The results of the remote status command.

        """
        return self.host.run_sync(f"/usr/bin/systemctl status {self._service}", is_safe=True)

    def start_mysql(self) -> Iterator[tuple[NodeSet, MsgTreeElem]]:
        """Starts mariadb service.

        Returns:
            The results of the remote start command.

        """
        return self.host.run_sync(f"/usr/bin/systemctl start {self._service}", print_output=True)

    def restart_mysql(self) -> Iterator[tuple[NodeSet, MsgTreeElem]]:
        """Restarts mariadb service.

        Returns:
            The results of the remote restart command.

        """
        return self.host.run_sync(f"/usr/bin/systemctl restart {self._service}")

    def clean_data_dir(self, *, skip_confirmation: bool = False) -> None:
        """Removes everything contained in the data directory.

        Arguments:
            skip_confirmation: execute the operation without any user confirmation.

        """
        command = f"/usr/bin/rm -rf {self._data_dir}"
        if not skip_confirmation:
            ask_confirmation(f"ATTENTION: destructive action for {self.host}: {command}. Are you sure to proceed?")

        self.host.run_sync(command)

    def upgrade(self) -> Iterator[tuple[NodeSet, MsgTreeElem]]:
        """Runs the relevant mysql_upgrade command to upgrade the instance content.

        Returns:
            The results of the remote upgrade command.

        """
        return self.host.run_sync(f"/usr/local/bin/mysql_upgrade --socket {self._sock} --force", print_output=True)

    def get_replication_info(self) -> ReplicationInfo:
        """Get the replication information suitable to set a new node's replication.

        Returns:
            The replication information object, useful to setup a new instance's replication to resume from the same
            position.

        """
        replication_status = self.show_slave_status()
        info = ReplicationInfo(
            primary=replication_status.get("Master_Host", ""),
            binlog=replication_status.get("Master_Log_File", ""),
            position=int(replication_status.get("Exec_Master_Log_Pos", -1)),
            port=int(replication_status.get("Master_Port", -1)),
        )
        if not (info.primary and info.binlog and info.position > -1 and info.port > -1):
            raise MysqlLegacyError(f"Could not find the replication position: {info}")

        logger.debug("Replication info for %s: %s", self.host, info)
        return info

    @property
    def primary(self) -> str:
        """Retrieves the replication source of this cluster.

        Raises:
            spicerack.mysql_legacy.MysqlLegacyError: if unable to find the master host of the current instance.

        """
        if not self._primary:
            try:
                self._primary = self.show_slave_status()["Master_Host"]
            except (KeyError, MysqlLegacyError) as e:
                raise MysqlLegacyError("Unable to retrieve master host") from e

        return self._primary

    def prep_src_for_cloning(self) -> ReplicationInfo:
        """Helper that prepares source instance to be cloned.

        Returns:
            The replication information object, useful to setup a new instance's replication to resume from the same
            position.

        """
        self.stop_slave()
        replication_info = self.get_replication_info()
        self.stop_mysql()
        return replication_info

    def set_replication_parameters(self, *, replication_info: ReplicationInfo, user: str, password: str) -> None:
        """Sets the replication parameters for the MySQL instance."""
        query = f"""
            CHANGE MASTER TO master_host='{replication_info.primary}',
            master_port={replication_info.port},
            master_ssl=1,
            master_log_file='{replication_info.binlog}',
            master_log_pos={replication_info.position},
            master_user='{user}',
            master_password='{password}'
        """
        self.run_query(dedent(query).strip())

    def post_clone_reset_with_slave_stopped(self) -> None:
        """Prepares the MySQL instance for the first pooling operation."""
        self.host.run_sync(
            f"chown -R mysql:mysql {self._data_dir}",
            '/usr/bin/systemctl set-environment MYSQLD_OPTS="--skip-slave-start"',
        )
        self.start_mysql()
        self.stop_slave()
        self.run_query("RESET SLAVE ALL")

    def resume_replication(self) -> None:
        """Resumes replication on the source MySQL instance."""
        self.host.run_sync('/usr/bin/systemctl set-environment MYSQLD_OPTS="--skip-slave-start"')
        self.start_mysql()
        self.upgrade()
        self.restart_mysql()
        self.start_slave()

    @retry(
        tries=480,  # We allow up to 8 hours for replication lag to catch up
        delay=timedelta(seconds=60),
        backoff_mode="constant",
        exceptions=(MysqlLegacyReplagError,),
    )
    def wait_for_replication(self, threshold: float = 1.0) -> None:
        """Waits for replication to catch up.

        Arguments:
            threshold: the replication lag threshold in seconds under which the replication is considered in sync.

        Raises:
            spicerack.mysql_legacy.MysqlLegacyReplagError: if the replication lag is still too high after all the
                retries.

        """
        replag = self.replication_lag()
        if replag > threshold:
            raise MysqlLegacyReplagError(f"Replication lag higher than the threshold ({threshold}s): {replag}s")

    def replication_lag(self) -> float:
        """Retrieves the current replication lag.

        Returns:
            The replication lag in seconds.

        Raises:
            spicerack.mysql_legacy.MysqlLegacyError: if no lag information is present or unable to parse the it.

        """
        query = """
            SELECT greatest(0, TIMESTAMPDIFF(MICROSECOND, max(ts), UTC_TIMESTAMP(6)) - 500000)/1000000 AS lag
            FROM heartbeat.heartbeat
            ORDER BY ts LIMIT 1
        """
        query = dedent(query).strip()
        query_res = list(self.run_query(query, is_safe=True))
        if not query_res:
            raise MysqlLegacyError("Got no output from the replication lag query")

        output = ""
        try:
            output = query_res[0][1].message().decode("utf-8").splitlines()
            return float(output[1])
        except (IndexError, ValueError) as e:
            raise MysqlLegacyError(f"Unable to parse replication lag from: {output}") from e


class MysqlLegacyRemoteHosts(RemoteHostsAdapter):
    """Custom RemoteHosts class for executing MySQL queries."""

    # TODO merge this method with Instance.run_query()
    def run_query(self, query: str, database: str = "", **kwargs: Any) -> Iterator[tuple[NodeSet, MsgTreeElem]]:
        """Execute the query via Remote.

        Arguments:
            query: the mysql query to be executed. Double quotes must be already escaped.
            database: an optional MySQL database to connect to before executing the query.
            **kwargs: any additional argument is passed to :py:meth:`spicerack.remote.RemoteHosts.run_sync`. By default
                the ``print_progress_bars`` and ``print_output`` arguments are set to :py:data:`False`.

        Raises:
            spicerack.remote.RemoteExecutionError: if the Cumin execution returns a non-zero exit code.

        """
        command = f'/usr/local/bin/mysql --skip-ssl --skip-column-names --batch -e "{query}" {database}'.strip()
        kwargs.setdefault("print_progress_bars", False)
        kwargs.setdefault("print_output", False)
        return self._remote_hosts.run_sync(command, **kwargs)

    def list_hosts_instances(self, *, group: bool = False) -> list[Instance]:
        """List MariaDB instances on the host.

        Arguments:
            group: not yet implemented feature to allow parallelization.

        Raises:
            spicerack.remote.RemoteExecutionError: if the Cumin execution returns a non-zero exit code.
            NotImplementedError: if the replag is not fully caught on.

        """
        if len(self._remote_hosts) != 1:
            raise NotImplementedError("Only single host are supported at this time.")

        if group:
            # TODO see this comment:
            # https://gerrit.wikimedia.org/r/c/operations/software/spicerack/+/1005531/comment/7af929a5_6d6184d4/
            # we could use this method to parallelize stuff on instances as well.
            raise NotImplementedError("Grouping and parallelization are not supported at this time.")

        return self._list_host_instances()

    def _list_host_instances(self) -> list[Instance]:
        """List MariaDB instances on the host.

        Raises:
            spicerack.remote.RemoteExecutionError: if the Cumin execution returns a non-zero exit code.

        """
        instances: list[Instance] = []
        command = "/usr/bin/systemctl --no-pager --type=service --plain --no-legend  list-units 'mariadb*'"
        service_list = list(
            self._remote_hosts.run_sync(command, is_safe=True, print_progress_bars=False, print_output=False)
        )
        if not service_list:
            return instances

        services = service_list[0][1].message().decode("utf8").splitlines()
        if len(services) == 1 and services[0].split()[0] == "mariadb.service":
            instances.append(Instance(self._remote_hosts))
            return instances

        for service in services:
            service_name = service.split()[0]
            if service_name.startswith("mariadb@") and service_name.endswith(".service"):
                instances.append(Instance(self._remote_hosts, name=service_name[8:-8]))

        return instances


class MysqlLegacy:
    """Class to manage MySQL servers."""

    # FIXME this query could be replaced by the one in _get_replication() as it's the one used in monitoring
    heartbeat_query: str = (
        "SELECT ts FROM heartbeat.heartbeat WHERE datacenter = '{dc}' and shard = '{section}' "
        "ORDER BY ts DESC LIMIT 1"
    )
    """Query pattern to check the heartbeat for a given datacenter and section."""

    def __init__(self, remote: Remote, dry_run: bool = True) -> None:
        """Initialize the instance.

        Arguments:
            remote: the Remote instance.
            dry_run: whether this is a DRY-RUN.

        """
        self._remote = remote
        self._dry_run = dry_run

    def get_dbs(self, query: str) -> MysqlLegacyRemoteHosts:
        """Get a MysqlLegacyRemoteHosts instance for the matching hosts.

        Arguments:
            query: the Remote query to use to fetch the DB hosts.

        """
        return MysqlLegacyRemoteHosts(self._remote.query(query))

    def get_core_dbs(
        self,
        *,
        datacenter: Optional[str] = None,
        section: Optional[str] = None,
        replication_role: Optional[str] = None,
        excludes: tuple[str, ...] = (),
    ) -> MysqlLegacyRemoteHosts:
        """Get an instance to operated on the core databases matching the parameters.

        Arguments:
            datacenter: the name of the datacenter to filter for, accepted values are those specified in
                :py:data:`spicerack.constants.CORE_DATACENTERS`.
            replication_role: the repication role to filter for, accepted values are those specified in
                :py:data:`spicerack.mysql_legacy.REPLICATION_ROLES`.
            section: a specific section to filter for, accepted values are those specified in
                :py:data:`spicerack.mysql_legacy.CORE_SECTIONS`.
            excludes: sections to exclude from getting.

        Raises:
            spicerack.mysql_legacy.MysqlLegacyError: on invalid data or unexpected matching hosts.

        """
        query_parts = ["A:db-core"]
        dc_multipler = len(CORE_DATACENTERS)
        section_multiplier = len(CORE_SECTIONS)

        if datacenter is not None:
            dc_multipler = 1
            if datacenter not in CORE_DATACENTERS:
                raise MysqlLegacyError(f"Got invalid datacenter {datacenter}, accepted values are: {CORE_DATACENTERS}")

            query_parts.append("A:" + datacenter)

        for exclude in excludes:
            if exclude not in CORE_SECTIONS:
                raise MysqlLegacyError(f"Got invalid excludes {exclude}, accepted values are: {CORE_SECTIONS}")
            section_multiplier -= 1
            query_parts.append(f"not A:db-section-{exclude}")

        if section is not None:
            section_multiplier = 1
            if section not in CORE_SECTIONS:
                raise MysqlLegacyError(f"Got invalid section {section}, accepted values are: {CORE_SECTIONS}")

            query_parts.append(f"A:db-section-{section}")

        if replication_role is not None:
            if replication_role not in REPLICATION_ROLES:
                raise MysqlLegacyError(
                    f"Got invalid replication_role {replication_role}, accepted values are: {REPLICATION_ROLES}"
                )

            query_parts.append(f"A:db-role-{replication_role}")

        mysql_hosts = MysqlLegacyRemoteHosts(self._remote.query(" and ".join(query_parts)))

        # Sanity check of matched hosts in case of master selection
        if replication_role == "master" and len(mysql_hosts) != dc_multipler * section_multiplier:
            raise MysqlLegacyError(f"Matched {len(mysql_hosts)} masters, expected {dc_multipler * section_multiplier}")

        return mysql_hosts

    def set_core_masters_readonly(self, datacenter: str) -> None:
        """Set the core masters in read-only.

        Arguments:
            datacenter: the name of the datacenter to filter for.

        Raises:
            spicerack.remote.RemoteExecutionError: on Remote failures.
            spicerack.mysql_legacy.MysqlLegacyError: on failing to verify the modified value.

        """
        logger.debug("Setting core DB masters in %s to be read-only", datacenter)
        target = self.get_core_dbs(datacenter=datacenter, replication_role="master")
        target.run_query("SET GLOBAL read_only=1")
        self.verify_core_masters_readonly(datacenter, True)

    def set_core_masters_readwrite(self, datacenter: str) -> None:
        """Set the core masters in read-write.

        Arguments:
            datacenter: the name of the datacenter to filter for.

        Raises:
            spicerack.remote.RemoteExecutionError: on Remote failures.
            spicerack.mysql_legacy.MysqlLegacyError: on failing to verify the modified value.

        """
        logger.debug("Setting core DB masters in %s to be read-write", datacenter)
        target = self.get_core_dbs(datacenter=datacenter, replication_role="master")
        target.run_query("SET GLOBAL read_only=0")
        self.verify_core_masters_readonly(datacenter, False)

    def verify_core_masters_readonly(self, datacenter: str, is_read_only: bool) -> None:
        """Verify that the core masters are in read-only or read-write mode.

        Arguments:
            datacenter: the name of the datacenter to filter for.
            is_read_only: whether the read-only mode should be set or not.

        Raises:
            spicerack.mysql_legacy.MysqlLegacyError: on failure.

        """
        logger.debug(
            "Verifying core DB masters in %s have read-only=%d",
            datacenter,
            is_read_only,
        )
        target = self.get_core_dbs(datacenter=datacenter, replication_role="master")
        expected = str(int(is_read_only))  # Convert it to the returned value from MySQL: 1 or 0.
        failed = False

        for nodeset, output in target.run_query("SELECT @@global.read_only", is_safe=True):
            response = output.message().decode()
            if response != expected:
                logger.error(
                    "Expected output to be '%s', got '%s' for hosts %s",
                    expected,
                    response,
                    str(nodeset),
                )
                failed = True

        if failed and not self._dry_run:
            raise MysqlLegacyError(
                f"Verification failed that core DB masters in {datacenter} have read-only={is_read_only}"
            )

    def check_core_masters_in_sync(self, dc_from: str, dc_to: str) -> None:
        """Check that all core masters in dc_to are in sync with the core masters in dc_from.

        Arguments:
            dc_from: the name of the datacenter from where to get the master positions.
            dc_to: the name of the datacenter where to check that they are in sync.

        Raises:
            spicerack.remote.RemoteExecutionError: on failure.

        """
        logger.debug("Waiting for the core DB masters in %s to catch up", dc_to)
        heartbeats = self.get_core_masters_heartbeats(dc_from, dc_from)
        self.check_core_masters_heartbeats(dc_to, dc_from, heartbeats)

    def get_core_masters_heartbeats(self, datacenter: str, heartbeat_dc: str) -> dict[str, datetime]:
        """Get the current heartbeat values from core DB masters in DC for a given heartbeat DC.

        Arguments:
            datacenter: the name of the datacenter from where to get the heartbeat values.
            heartbeat_dc: the name of the datacenter for which to filter the heartbeat query.

        Returns:
            A dictionary with the section name :py:class:`str` as keys and their heartbeat
            :py:class:`datetime.datetime` as values. For example::

                {'s1': datetime.datetime(2018, 1, 2, 11, 22, 33, 123456)}

        Raises:
            spicerack.mysql_legacy.MysqlLegacyError: on failure to gather the heartbeat or convert it into a datetime.

        """
        heartbeats = {}
        for section in CORE_SECTIONS:
            core_dbs = self.get_core_dbs(datacenter=datacenter, section=section, replication_role="master")
            heartbeats[section] = MysqlLegacy._get_heartbeat(core_dbs, section, heartbeat_dc)

        return heartbeats

    def check_core_masters_heartbeats(
        self, datacenter: str, heartbeat_dc: str, heartbeats: dict[str, datetime]
    ) -> None:
        """Check the current heartbeat values in the core DB masters in DC are in sync with the provided heartbeats.

        Arguments:
            datacenter: the name of the datacenter from where to get the heartbeat values.
            heartbeat_dc: the name of the datacenter for which to filter the heartbeat query.
            heartbeats: a dictionary with the section name :py:class:`str` as keys and heartbeat
                :py:class:`datetime.datetime` for each core section as values.

        Raises:
            spicerack.mysql_legacy.MysqlLegacyError: on failure to gather the heartbeat or convert it into a datetime.

        """
        for section, heartbeat in heartbeats.items():
            self._check_core_master_in_sync(datacenter, heartbeat_dc, section, heartbeat)

    @retry(exceptions=(MysqlLegacyError,))
    def _check_core_master_in_sync(
        self,
        datacenter: str,
        heartbeat_dc: str,
        section: str,
        parent_heartbeat: datetime,
    ) -> None:
        """Check and retry that the heartbeat value in a core DB master in DC is in sync with the provided heartbeat.

        Arguments:
            datacenter: the name of the datacenter from where to get the heartbeat value.
            heartbeat_dc: the name of the datacenter for which to filter the heartbeat query.
            section: the section name from where to get the heartbeat value and filter the heartbeat query.
            master_heartbeat: the reference heartbeat from the parent master to use to verify this master is in sync
                with it.

        Raises:
            spicerack.mysql_legacy.MysqlLegacyError: on failure to gather the heartbeat or convert it into a datetime
                or not yet in sync.

        """
        core_dbs = self.get_core_dbs(datacenter=datacenter, section=section, replication_role="master")
        local_heartbeat = MysqlLegacy._get_heartbeat(core_dbs, section, heartbeat_dc)

        # The check requires that local_heartbeat is stricly greater than parent_heartbeat because heartbeat writes also
        # when the DB is in read-only mode and has a granularity of 1s (as of 2018-09), meaning that an event could have
        # been written after the last heartbeat but before the DB was set in read-only mode and that event could not
        # have been replicated, hence checking the next heartbeat to ensure they are in sync.
        if local_heartbeat <= parent_heartbeat:
            delta = (local_heartbeat - parent_heartbeat).total_seconds()
            raise MysqlLegacyError(
                f"Heartbeat from master {core_dbs} for section {section} not yet in sync: "
                f"{local_heartbeat} <= {parent_heartbeat} (delta={delta})"
            )

    @staticmethod
    def _get_heartbeat(mysql_hosts: MysqlLegacyRemoteHosts, section: str, heartbeat_dc: str) -> datetime:
        """Get the heartbeat from the remote host for a given DC.

        Arguments:
            mysql_hosts: the instance for the target DB to query.
            section: the DB section for which to get the heartbeat.
            heartbeat_dc: the name of the datacenter for which to filter the heartbeat query.

        Raises:
            spicerack.mysql_legacy.MysqlLegacyError: on failure to gather the heartbeat or convert it into a datetime.

        """
        query = MysqlLegacy.heartbeat_query.format(dc=heartbeat_dc, section=section)

        for _, output in mysql_hosts.run_query(query, is_safe=True):
            try:
                heartbeat_str = output.message().decode()
                heartbeat = datetime.strptime(heartbeat_str, "%Y-%m-%dT%H:%M:%S.%f")
                break
            except (TypeError, ValueError) as e:
                raise MysqlLegacyError(f"Unable to convert heartbeat '{heartbeat_str}' into datetime") from e
        else:
            raise MysqlLegacyError(f"Unable to get heartbeat from master {mysql_hosts} for section {section}")

        return heartbeat



"""Cookbook internal module."""
import argparse
import importlib
import logging
import os
import sys
from collections.abc import Sequence
from pathlib import Path
from typing import Optional, cast

from wmflib.config import load_yaml_config

from spicerack import Spicerack, SpicerackExtenderBase, _log, _module_api, cookbook
from spicerack._menu import BaseItem, CookbookItem, MenuError, TreeItem
from spicerack.exceptions import SpicerackError

logger = logging.getLogger(__name__)


class CookbookError(SpicerackError):
    """Custom exception class for errors of this module."""


class CookbookCollection:
    """Collect and represent available cookbooks."""

    cookbooks_module_prefix: str = "cookbooks"

    def __init__(
        self,
        *,
        base_dirs: list[Path],
        args: Sequence[str],
        spicerack: Spicerack,
        path_filter: str = "",
    ) -> None:
        """Initialize the class and collect all the cookbook menu items.

        Arguments:
            base_dirs: the list of base directories from where to start looking for cookbooks.
            args: the list of arguments to pass to the collected items.
            spicerack: the initialized instance of the library.
            path_filter: an optional relative module path to filter for. If set, only cookbooks that are part of this
                subtree will be collected.

        """
        self.base_dirs = [base_dir / self.cookbooks_module_prefix for base_dir in base_dirs]
        self.args = args
        self.spicerack = spicerack
        self.path_filter = path_filter

        module = import_module(self.cookbooks_module_prefix)
        self.menu = TreeItem(module, self.args, self.spicerack, self.cookbooks_module_prefix)
        for base_dir in self.base_dirs:
            self._collect(base_dir)

    def get_item(self, path: str) -> Optional[BaseItem]:
        """Retrieve the item for a given path.

        Arguments:
            path: the path of the item to look for.

        Returns:
            None: when no item is found.
            spicerack._menu.TreeItem: when the item found is a menu of cookbooks.
            spicerack._menu.CookbookItem: when the item found is a single cookbook.

        """
        if path:
            effective_path = ".".join((self.cookbooks_module_prefix, path))
        else:
            effective_path = self.cookbooks_module_prefix

        item: Optional[BaseItem] = None
        parts = effective_path.split(".")
        multiple_class_name = ".".join(parts[-2:])
        for i, subpath in enumerate(parts):
            if i == 0:  # Initial menu
                item = self.menu
            elif item is not None and subpath in item.items.keys():  # Submenu or child cookbook found
                item = item.items[subpath]
            elif item is not None and i == len(parts) - 2 and multiple_class_name in item.items.keys():
                # Found cookbook in a module that has multiple class API cookbooks
                item = item.items[multiple_class_name]
                break
            else:
                item = None

        return item

    def _create_menu_for_path(self, path: str) -> TreeItem:
        """Create the menu for a given path, including intermediate levels, if missing. Return the existing one if any.

        Arguments:
            path: the path of the item to look for.

        """
        item: TreeItem = self.menu
        if path == self.cookbooks_module_prefix:
            return item

        progressive_path = [self.cookbooks_module_prefix]
        for subpath in path.split(".")[1:]:
            progressive_path.append(subpath)
            if subpath in item.items:
                item = cast(TreeItem, item.items[subpath])
            else:
                module_name = ".".join(progressive_path)
                submenu = TreeItem(
                    import_module(module_name),
                    self.args,
                    self.spicerack,
                    self.cookbooks_module_prefix,
                )
                # When collecting the cookbooks and creating the TreeItem instances, the relation to the parent
                # menu should be skipped for those intermediate menus created for coherence but that should not be
                # accessible by the user, like when using a path_filter.
                add_parent = submenu.full_name.startswith(self.path_filter) and submenu.full_name != self.path_filter
                item.append(submenu, add_parent=add_parent)
                item = submenu

        return item

    def _collect(self, base_dir: Path) -> None:
        """Collect available cookbooks starting from a base path.

        Arguments:
            base_dir: the directory where to start collecting the cookbooks.

        """
        for dirpath in base_dir.rglob(""):  # Selects only directories
            if dirpath.name == "__pycache__":
                continue

            relpath = dirpath.relative_to(base_dir)
            if relpath.name:
                prefix = str(relpath).replace("/", ".").rstrip(".")
                module_prefix = f"{self.cookbooks_module_prefix}.{prefix}"
            else:
                prefix = ""
                module_prefix = self.cookbooks_module_prefix

            if self._should_filter(prefix):
                continue

            try:
                menu = self._create_menu_for_path(module_prefix)
            except CookbookError as e:
                logger.error(e)
                continue

            for filepath in dirpath.glob("[!_]*.py"):  # Excludes files starting with an underscore, like __init__.py
                module_name = f"{module_prefix}.{filepath.stem}"
                self._collect_filename(module_name, menu)

    def _collect_filename(self, module_name: str, menu: TreeItem) -> None:
        """Collect all the available cookbooks in the given module and add them to the menu.

        Arguments:
            module_name: the Python module to load.
            menu: the menu to append the collected cookbook to.

        """
        try:
            classes = self._collect_module_cookbooks(import_module(module_name))
        except CookbookError as e:
            logger.error(e)
            return

        for class_obj in classes:
            try:
                cookbook_item = CookbookItem(class_obj, self.args, self.spicerack)
            except MenuError as e:
                logger.error(e)
                continue

            if self._should_filter(cookbook_item.full_name):
                continue

            menu.append(cookbook_item)

    def _should_filter(self, name: str) -> bool:
        """Check if a given path or full name should be skipped because not matching the current filter.

        Arguments:
            name: the name to check, can be either a CookbookItem full name or a TreeItem path.

        Returns:
            :py:data:`True` if the item should be skipped, :py:data:`False` otherwise.

        """
        if not self.path_filter:
            return False

        if name.startswith(self.path_filter[: len(name)]):
            return False

        return True

    def _collect_module_cookbooks(
        self, module: _module_api.CookbooksModuleInterface
    ) -> list[type[cookbook.CookbookBase]]:
        """Collect all classes derived from CookbookBase in the given module.

        Arguments:
            module: the module to check for cookbook classes.

        Returns:
            list: a list of CookbookBase derived class objects.

        """
        attrs = [getattr(module, attr) for attr in dir(module)]
        classes = [
            attr
            for attr in attrs
            if isinstance(attr, type) and issubclass(attr, cookbook.CookbookBase) and attr.__module__ == module.__name__
        ]

        if not classes:  # No class API cookbook found, convert a module API cookbook into a class
            classes.append(self._convert_module_in_cookbook(module))
        elif len(classes) == 1:  # Avoid the unnecessary further nested namespace
            class_obj = classes[0]
            full_name = class_obj.__module__.split(".", 1)[1]
            if "." in full_name:
                class_obj.spicerack_path, class_obj.spicerack_name = full_name.rsplit(".", 1)
            else:
                class_obj.spicerack_path = ""
                class_obj.spicerack_name = full_name
        else:  # Inject the class module name as part of the class name and set the module as that of the parent
            for class_obj in classes:
                full_name_prefix = class_obj.__module__.split(".", 1)[1]
                if "." in full_name_prefix:
                    class_obj.spicerack_path, name_prefix = full_name_prefix.rsplit(".", 1)
                    class_obj.spicerack_name = ".".join([name_prefix, class_obj.__name__])
                else:
                    class_obj.spicerack_path = ""
                    class_obj.spicerack_name = ".".join([full_name_prefix, class_obj.__name__])

        return classes

    def _convert_module_in_cookbook(self, module: _module_api.CookbooksModuleInterface) -> type[cookbook.CookbookBase]:
        """Convert a module API based cookbook into a class API cookbook.

        Arguments:
            module: the module to convert.

        Returns:
            type: a dynamically generated class derived from :py:class:`spicerack.cookbook.CookbookBase`.

        """
        module_name, name = module.__name__.rsplit(".", 1)
        try:
            title = module.__title__.splitlines()[0]  # Force it to be one-line only
        except AttributeError as e:
            logger.debug("Unable to detect title for module %s: %s", module.__name__, e)
            title = CookbookItem.fallback_title

        try:
            run = module.run
        except AttributeError as e:
            raise CookbookError(f"Unable to find run function in module {module.__name__}") from e

        if hasattr(module, "MAX_CONCURRENCY"):
            max_concurrency = module.MAX_CONCURRENCY
        else:
            max_concurrency = cookbook.CookbookRunnerBase.max_concurrency

        if hasattr(module, "LOCK_TTL"):
            lock_ttl = module.LOCK_TTL
        else:
            lock_ttl = cookbook.CookbookRunnerBase.lock_ttl

        runner_name = f"{name}Runner"
        runner = type(
            runner_name,
            (_module_api.CookbookModuleRunnerBase,),
            {
                "_run": staticmethod(run),
                "max_concurrency": max_concurrency,
                "lock_ttl": lock_ttl,
            },
        )
        runner.__module__ = module_name

        attributes = {
            "__module__": module_name,
            "__name__": name,
            "spicerack_name": name,
            "spicerack_path": module_name.split(".", 1)[1] if "." in module_name else "",
            "title": title,
            "get_runner": lambda _, args: runner(args, self.spicerack),
        }

        try:
            args_parser = module.argument_parser
            attributes["argument_parser"] = lambda _: args_parser()
        except AttributeError:
            # The cookbook doesn't accept any argument, set a default empty parser
            attributes["argument_parser"] = lambda _: argparse.ArgumentParser(
                description=module.__doc__, formatter_class=cookbook.ArgparseFormatter
            )

        cookbook_class = type(name, (cookbook.CookbookBase,), attributes)
        return cookbook_class


def argument_parser() -> argparse.ArgumentParser:
    """Get the CLI argument parser.

    If the COOKBOOK is passed as a path, it will be converted to a Python module syntax.

    """
    parser = argparse.ArgumentParser(
        description="Spicerack Cookbook Runner",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        epilog="See also https://wikitech.wikimedia.org/wiki/Spicerack/Cookbooks",
        allow_abbrev=False,  # Prevent matching of cookbook options
    )
    parser.add_argument(
        "-l",
        "--list",
        action="store_true",
        help=(
            "List all available cookbooks, if -v/--verbose is set print also their description. If a COOKBOOK is "
            "also specified, it will be used as a prefix filter."
        ),
    )
    parser.add_argument(
        "-c",
        "--config-file",
        default="/etc/spicerack/config.yaml",
        help="Path to the Spicerack configuration file to load.",
    )
    parser.add_argument(
        "-d",
        "--dry-run",
        action="store_true",
        help="Set the DRY-RUN mode, also for the cookbook.",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="Verbose output, also for the cookbook.",
    )
    parser.add_argument(
        "--no-locks",
        action="store_true",
        help="Do not acquire or check distributed locks on etcd. To be used only if there are issues with etcd.",
    )
    parser.add_argument(
        "cookbook",
        metavar="COOKBOOK",
        nargs="?",
        type=cookbook_path_type,
        default="",
        help=(
            "Either the name of the Python module to execute (sre.hosts.downtime) or the relative path of the Python "
            "file to execute (sre/hosts/downtime.py).  If the selected path/module is a directory (sre.hosts) or is "
            "not set, an interactive menu will be shown. Each directory listed in the `cookbooks_base_dirs` key of the"
            "configuration file represents a checked out cookbook repository containing a `cookbooks` subfolder. All "
            "paths specified must be relative to theses subfolders."
        ),
    )
    parser.add_argument(
        "cookbook_args",
        metavar="COOKBOOK_ARGS",
        nargs=argparse.REMAINDER,
        help="Collect all the remaining arguments to be passed to the cookbook or menu to execute.",
    )

    return parser


def cookbook_path_type(path: str) -> str:
    """Convert a COOKBOOK path to module syntax, if it's in path syntax.

    Arguments:
        path: the path to be converted.

    """
    if path.endswith(".py"):
        path = path[:-3].replace("/", ".")

    return path


def import_module(module_name: str) -> _module_api.CookbooksModuleInterface:
    """Import a Python module and return it.

    Arguments:
        module_name: the name of the module to load.

    Raises:
        spicerack._cookbook.CookbookError: on failure to load the module.

    """
    try:
        module = importlib.import_module(module_name)
    except Exception as e:
        raise CookbookError(f"Failed to import module {module_name}: {e}") from e

    return cast(_module_api.CookbooksModuleInterface, module)


def _import_extender_class(extender_class_param: str) -> SpicerackExtenderBase:
    """Import the extender class, ensure it's a subclass of SpicerackExtenderBase and return it.

    Arguments:
        extender_class_param: the configuration file parameter with the fully qualified class name.

    """
    extender_module_name, extender_class_name = extender_class_param.rsplit(".", 1)
    extender_module = import_module(extender_module_name)
    extender_class = getattr(extender_module, extender_class_name)
    if issubclass(extender_class, SpicerackExtenderBase):
        return extender_class

    raise TypeError(f"Extender class {extender_class_param} is not a subclass of spicerack.SpicerackExtenderBase.")


def main(argv: Optional[Sequence[str]] = None) -> Optional[int]:  # noqa: MC0001
    """Entry point, run the tool.

    Arguments:
        argv: the list of command line arguments to parse.

    Returns:
        The return code, zero on success, non-zero on failure.

    """
    args = argument_parser().parse_args(argv)
    if not args.cookbook:
        args.cookbook = ""

    config = load_yaml_config(args.config_file)
    cookbooks_base_dirs = []
    for base_dir in config["cookbooks_base_dirs"]:
        base_dir_path = Path(base_dir).expanduser()
        cookbooks_base_dirs.append(base_dir_path)
        sys.path.append(str(base_dir_path))

    if not cookbooks_base_dirs:
        print(
            "No cookbooks paths are specified in the `cookbooks_base_dirs` key of the configuration file.",
            file=sys.stderr,
        )
        return 1

    if config.get("external_modules_dir") is not None:
        sys.path.append(str(Path(config["external_modules_dir"]).expanduser()))

    def get_cookbook(spicerack: Spicerack, cookbook_path: str, cookbook_args: Sequence[str] = ()) -> Optional[BaseItem]:
        """Run a single cookbook.

        Arguments:
            argv: a sequence of strings of command line arguments to parse.

        Returns:
            :py:data:`None` on success, the return code otherwise, zero on success, non-zero on failure.

        """
        cookbooks = CookbookCollection(
            base_dirs=cookbooks_base_dirs,
            args=cookbook_args,
            spicerack=spicerack,
            path_filter=cookbook_path,
        )
        return cookbooks.get_item(cookbook_path)

    params = config.get("instance_params", {})
    params.update({"verbose": args.verbose, "dry_run": args.dry_run, "get_cookbook_callback": get_cookbook})
    if "extender_class" in params:
        try:
            params["extender_class"] = _import_extender_class(params["extender_class"])
        except Exception as e:  # pylint: disable=broad-except
            print(f"Failed to import the extender_class {params['extender_class']}:", e, file=sys.stderr)
            return 1

    if args.no_locks:
        params["etcd_config"] = ""  # Disable locking support

    try:
        spicerack = Spicerack(**params)
    except TypeError as e:
        print("Unable to instantiate Spicerack, check your configuration:", e, file=sys.stderr)
        return 1

    cookbooks = CookbookCollection(
        base_dirs=cookbooks_base_dirs, args=args.cookbook_args, spicerack=spicerack, path_filter=args.cookbook
    )
    if args.list:
        print(cookbooks.menu.get_tree(), end="")
        return 0

    cookbook_item = cookbooks.get_item(args.cookbook)
    if cookbook_item is None:
        logger.error("Unable to find cookbook %s", args.cookbook)
        return cookbook.NOT_FOUND_RETCODE

    base_path = Path(config["logs_base_dir"]).expanduser() / cookbook_item.path.replace(".", os.sep)
    _log.setup_logging(
        base_path,
        cookbook_item.name,
        spicerack.username,
        dry_run=args.dry_run,
        host=config.get("tcpircbot_host", None),
        port=int(config.get("tcpircbot_port", 0)),
    )

    logger.debug("Executing cookbook %s with args: %s", args.cookbook, args.cookbook_args)
    return cookbook_item.run()



"""Interactive module tests."""
from unittest import mock

import pytest

from spicerack import interactive
from spicerack.exceptions import SpicerackError


def test_get_management_password_in_env(monkeypatch):
    """Should return the MGMT_PASSWORD environment variable value, if set."""
    monkeypatch.setenv("MGMT_PASSWORD", "env_password")
    assert interactive.get_management_password() == "env_password"


@mock.patch("wmflib.interactive.getpass")
def test_get_management_password_interactive(mocked_getpass, monkeypatch):
    """Should ask for the password the MGMT_PASSWORD environment variable value, if set."""
    monkeypatch.delenv("MGMT_PASSWORD", raising=False)
    mocked_getpass.getpass.return_value = "interactive_password"
    assert interactive.get_management_password() == "interactive_password"
    mocked_getpass.getpass.assert_called_once_with(prompt="Management Password: ")


def test_get_management_password_empty(monkeypatch):
    """Should raise SpicerackError if the password is empty."""
    monkeypatch.setenv("MGMT_PASSWORD", "")
    with pytest.raises(SpicerackError, match="Empty Management Password"):
        interactive.get_management_password()
